# -*- coding: utf-8 -*-

import numpy as np
import math

# Mandatory TAU = 6.283185307179586...
math.tau = 2 * math.pi


def merge_where_nan(target, filler):
    """
    Procedurally mutates target by replacing its nan values by values from filler.
    This is a very simple polyfill for numpy.copyto (for numpy < 1.7).
    """
    try:
        np.copyto(target, filler, where=np.isnan(target))
    except AttributeError:
        isnan = np.isnan(target)
        target[isnan] = filler[isnan]


def safe_exp(x):
    """
    Should not throw during Underflows (and return 0) and Overflows (and return inf).
    This code is a hack, and may be cause of future bugs.
    Oddly enough, it passes a few basic tests in `math_utils_test.py`.
    """
    try:
        a = math.exp(x)
    except OverflowError:
        return float("inf")
    # There's no UnderflowError. WTF?
    # except UnderflowError:
    #     return 0
    return a


def divide(numerator, denominator, value_when_invalid=0):
    """
    Divide ndarray numerator by ndarray denominator, return the resulting ndarray.
    They MUST be of the same shape.
    This mutates numerator, too, of course.

    /!\ SLOW => UNUSED (kept as an iterator snippet)
    Benchmarking note: Devectorization of numpy one-liners is 30x slower.
    """

    it = np.nditer(numerator, flags=['multi_index'], op_flags=['readwrite'])
    while not it.finished:
        value = denominator[it.multi_index]
        if value != 0:
            it[0] /= value
        else:
            it[0] = value_when_invalid
        it.iternext()

    return numerator


def median_clip(data, clip_sigma=3., limit_ratio=1e-3, max_iterations=5):
    """
    Computes an iteratively sigma-clipped median on a data set.

    data : ndarray
        Input data.
    clip_sigma : float
        Sigma at which to clip.
    limit_ratio : float
        If the proportion of rejected pixels is less than this fraction, the iterations stop.
    max_iterations : int
        Ceiling on the number of clipping iterations.
    """

    # Make sure data is safe
    data = data[(np.isnan(data) == False) * np.isfinite(data)]

    median = np.median(data)
    iteration = 0
    finished = False
    while not finished:
        iteration += 1
        lastct = median
        median = np.median(data)
        sigma = np.std(data)

        # Reduce data set
        index = np.nonzero(np.abs(data - median) < clip_sigma * sigma)
        if np.size(index) > 0:
            data = data[index]

        if (abs(median - lastct) / abs(lastct) < limit_ratio) or (iteration >= max_iterations):
            finished = True

    median = np.median(data)
    sigma = np.std(data)

    return median, sigma, iteration

def flux_gaussian(radius, rhwhm):
        """
        Disk Gaussian profile with Gaussian thickness
        """
        energy = np.exp(-radius ** 2 / 2. / (2. * rhwhm / 2.35) ** 2) #* np.exp(-nz ** 2 / 2. / hz ** 2)
        return energy #DizkModel.flux_sersic(radius, size, 0.5)

def flux_sersic(radius, size, index):
        """
        Sersic profile with Gaussian thickness
        See http://en.wikipedia.org/wiki/Sersic_profile
        """
        beta = 1.9992 * index - 0.3271
        energy = np.exp(-beta * (radius / size) ** (1. / index))
        #energy = np.exp(-beta * (radius / size) ** (1. / index)) * np.exp(-nz ** 2 / 2. / hz ** 2)
        #energy = np.exp(-beta * (radius / size) ** (1. / index)) * np.exp(-np.sqrt(nz ** 2) / hz )  #exponential z-profile
        #energy = np.exp(-beta * (radius / size) ** (1. / index)) * np.cosh(-nz / hz )**(-2.)        #sech^2 z-profile
        return energy

def flux_exponential(radius, size):
        """
        Disk exponential profile with Gaussian thickness
        This is merely a Sérsic profile with index = 1
        """
        return flux_sersic(radius, size, 1.)

def flux_de_vaucouleurs(radius, size):
        """
        De Vaucouleurs profile with Gaussian thickness
        See http://en.wikipedia.org/wiki/De_Vaucouleurs%27_law
        This is merely a Sérsic profile with index = 4
        """
        return flux_sersic(radius, size, 4.)










# coding=utf-8

import logging
import os.path

from astropy.io import fits

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('FitsSanitizer')


# You can import this and use the two functions at your convenience
# or call this script : `python fits_sanitizer.py FILE`
# See options with `python fits_sanitizer.py -h`
# Note that you can use wildcards in the FILE.


def sanitize_fits_file(filename, prefix=''):
    fits_file = fits.open(filename)
    fits_file, fits_has_changed = sanitize_fits(fits_file)
    if fits_has_changed:
        prefix = str(prefix)
        filename = os.path.join(os.path.dirname(filename), prefix+os.path.basename(filename))
        fits_file.writeto(filename, clobber=True)


def sanitize_fits(hdulist):
    assert isinstance(hdulist, fits.HDUList)
    fits_has_changed = False
    for header_index, header in enumerate(hdulist):
        header = header.header
        for key in header:
            value = header.get(key)
            #print key+' -- '+str(value)
            changed = False

            # Lowercase 'DEG' to 'deg'
            if value == 'DEG':
                value = 'deg'
                changed = True
                logger.info("Lowercased %s 'DEG' into 'deg'.", key)

            if changed:
                fits_has_changed = True
                header.set(key, value)

        # Set um (microns) as the default CUNIT3 when missing
        #if not header.get('CUNIT3'):
        #    header.set('CUNIT3', 'um')

    return hdulist, fits_has_changed


# MAIN SCRIPT RUN WHEN LAUNCHED FROM CLI
if __name__ == '__main__':

    from argparse import ArgumentParser
    from argparse import RawTextHelpFormatter

    def is_valid_file(parser, arg):
        if not os.path.exists(arg):
            parser.error("The file %s does not exist!" % arg)
        else:
            return arg

    parser = ArgumentParser(formatter_class=RawTextHelpFormatter, description="""
Sanitize specified FITS files. By default, this will overwrite the FITS files.
To create another file(s), you can specify a prefix using the --prefix option.
What this actually does :
    - Lowercase 'DEG' unit
    - that's all folks !""")
    parser.add_argument('filenames', metavar='FILE', nargs='+',
                        type=lambda x: is_valid_file(parser, x),
                        help='A FITS file to sanitize')
    parser.add_argument('--prefix', dest='prefix', default='',
                        help='A prefix to prepend to the filename(s), to create new files')

    args = parser.parse_args()

    for filename in args.filenames:
        sanitize_fits_file(filename, args.prefix)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

#IMPORTS
import os,re
import sys

import scipy
from scipy import ndimage
from scipy.signal import convolve2d as conv2d_scipy
import astropy
import astropy.io.ascii as asciitable
from astropy.convolution import convolve_fft as conv2d_astro
from astropy.table import Table

import math
import numpy as np
np.random.seed(seed=1234)
from copy import deepcopy

# LOGGING CONFIGURATION
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK: Plots')

from .galaxy_parameters import GalaxyParameters
from .hyperspectral_cube import HyperspectralCube as HyperCube
from .convolution import convolve_2d as conv2d_intern

# OPTIONAL IMPORTS
try:
    import bottleneck as bn
except ImportError:
    logger.info("bottleneck (optional) not installed, performances will be degraded")
    import numpy as bn

#matplotlib
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import inset_locator as inl
# from matplotlib.ticker import MaxNLocator
from matplotlib import pyplot as plot


class Plots:

    def plot_mcmc(self, filename=None, method=None, plot_likelihood=False, adapt_range='5stdev', fontsize=10):
        """
        Plot the MCMC chain details, and then either show it or save it to a file.

        filepath: string
            If specified, will write the plot to a file instead of showing it.
            The file will be created at the provided filepath, be it absolute or relative.
            The extension of the file must be either png or pdf.
        plot_likelihood: bool
            True to plot -log[chi2] instead
        adapt_range: string
            'boundaries'  to adapt the range to boundaries
            'minmax' [default]  to adapt the range to min/max values
            '3stdev'   to adapt the range to 3 x stdev
            '5stdev'   to adapt the range to 5 x stdev
        fontsize: int
            to change the fontsize

        """
        if self.chain is None:
            raise RuntimeError(self.NO_CHAIN_ERROR)

        if filename is not None:
            name, extension = os.path.splitext(filename)
            supported_extensions = ['.png', '.pdf']
            if not extension in supported_extensions:
                raise ValueError("Extension '%s' is not supported, use %s.",
                                 extension, ', '.join(supported_extensions))

        if method is None:
            method = self.method

        if len(self.chain)>1000:
            fmt='b-'
        else:
            fmt='b.'
        alpha=1
        ms=3

        chain = self.chain.copy()
        xmin = self.chain.xmin
        xmax = self.chain.xmax
        chain_size = np.size(chain, 0)
        idx = np.arange(0,chain_size)

        if method == 'chi_sorted':
            chain.sort('reduced_chi')
            fmt='b.'
            ms=4
            alpha=0.4
            idx = np.random.randint(low=0, high=chain_size, size=1500)
            idx.sort()
            xmin = 0
            xmax = xmax * np.size(idx) / chain_size

        if fontsize is None:
            fontsize = plot.rcParams['font.size']

        fig = plot.figure(1, figsize=(16, 9))

        names = list(self.galaxy.names)
        names.append('reduced_chi')
        short_dict = self.galaxy.short_dict()

        plot.clf()  # clear current figure
        plot.subplots_adjust(wspace=0.32, hspace=0.32,
                             bottom=0.05, top=0.95, left=0.05, right=0.95)
        n = np.size(names)

        rows = 3
        cols = int(math.ceil((n-1) / (rows - 1.)))

        for i, par in enumerate(names):

            if i < n - 1:
                # Parameters
                ax=plot.subplot2grid((rows, cols), (int(math.floor(i / cols)), i % cols))
                if method == 'chi_sorted':
                    plot.plot(chain[par].data[idx], fmt, alpha=alpha, ms=ms)
                elif method == 'MAP':
                    idx_map = self.chain.idxsorted[0]
                    plot.plot(chain[par].data[idx], fmt, alpha=alpha, ms=ms, zorder=0)
                    plot.plot(idx_map, chain[par].data[idx_map], 'r+', alpha=alpha, ms=6, zorder=0)
                else:
                    plot.plot(chain[par].data[idx], fmt, alpha=alpha, ms=ms, zorder=0)
                plot.hlines(y=self.galaxy[i], xmin=xmin, xmax=xmax, color='r', label=r'$\hat P$', zorder=10)
                plot.hlines(y=self.galaxy[i] + self.galaxy.stdev[i], xmin=xmin, xmax=xmax, color='k',lw=2, label=r'$\sigma$', zorder=10)
                plot.hlines(y=self.galaxy[i] - self.galaxy.stdev[i], xmin=xmin, xmax=xmax, color='k',lw=2, zorder=10)
                if self.galaxy.lower is not None:
                    plot.hlines(y=self.galaxy.lower[i], xmin=xmin, xmax=xmax, color='k', linestyles='dotted',lw=2, zorder=10, label='%d ' % (self.percentile) + '% CI')
                if self.galaxy.upper is not None:
                    plot.hlines(y=self.galaxy.upper[i], xmin=xmin, xmax=xmax, color='k', linestyles='dotted',lw=2, zorder=10)

                #plot boundaries except for pa
                if 'pa' not in par:
                    try:
                        plot.axhline(self.min_boundaries[i],color='r',lw=1)
                        plot.axhline(self.max_boundaries[i],color='r',lw=1)
                    except:
                        self.logger.info("Boundaries not present, will not be shown")

                if i == n-2:
                    #plot.legend(loc=0, fontsize=13)
                    #plot.legend(loc=0, fontsize=13, bbox_to_anchor=(0.02, 0.3))
                    plot.legend(loc='lower center', fontsize=fontsize, bbox_to_anchor=(0.5, 0.0), ncol=3, fancybox=True, shadow=True)

                if (adapt_range == '5stdev') and (self.galaxy.stdev is not None):
                    plot.ylim(self.galaxy[i]-5.*self.galaxy.stdev[i], self.galaxy[i]+5.*self.galaxy.stdev[i])
                elif (adapt_range == '3stdev') and (self.galaxy.stdev is not None):
                    plot.ylim(self.galaxy[i]-3.*self.galaxy.stdev[i], self.galaxy[i]+3.*self.galaxy.stdev[i])
                elif (adapt_range =='boundaries') and (self.min_boundaries is not None and self.max_boundaries is not None):
                    plot.ylim(self.min_boundaries[i], self.max_boundaries[i])
                elif (adapt_range == 'minmax') or  (adapt_range is None):
                    plot.ylim(np.min(chain[par]), np.max(chain[par]) )

                title = r"%s" % (short_dict[par])
                title += "(%s)" % (self.galaxy.unit_dict()[par])
                plot.title(title, fontsize=fontsize)

            else:
                # Last row, reduced ki
                ax=plot.subplot2grid((rows, cols), (rows - 1, 0), colspan=cols)
                if (plot_likelihood):
                    ax.plot(np.exp(-chain['reduced_chi'] ))
                    ax.set_ylim([0,1])
                    ax.set_title(r'${\cal L}=\exp$[-$\chi^2$]')
                else:
                    ax.plot(np.log10(chain['reduced_chi'][idx] - np.min(chain['reduced_chi'])))
                    if method == 'MAP':
                        ax.axvline(idx_map,lw=1)
                    ax.set_title(r'$\log$ [$\chi^2 - \chi^2_{min}$]')

            if method is 'chi_sorted':
                ax.set_xticks([])
            else:
                ax.set_xticks(np.arange(2.) / 2 * np.size(idx))


        fig.subplots_adjust(wspace=0.3)

        if filename is None:
            plot.show()
        else:
            plot.savefig(filename)
            plot.close()

    def plot_geweke(self, filepath=None, fontsize=10, Nsigma=2, Nintervals=25, full_chain=False):
        """
        Plot the Geweke score for each parameter (from Geweke 1992: http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.27.2952)
         and then either show it or save it to a file.

        See also https://pymc-devs.github.io/pymc/modelchecking.html

        filepath: string
            If specified, will write the plot to a file instead of showing it.
            The file will be created at the provided filepath, be it absolute or relative.
            The extension of the file must be either png or pdf.
        fontsize: int [None default]
            to change the fontsize for plot
            if None will use rcParams
        Nsigma: int [3 default]
            sigma range for convergence test
        Nintervals
            Number of intervals for geweke statistics
        full_chain: boolean
            True if score for full chain

        """
        if self.chain is None:
            raise RuntimeError(self.NO_CHAIN_ERROR)

        if filepath is not None:
            name, extension = os.path.splitext(filepath)
            supported_extensions = ['.png', '.pdf']
            if not extension in supported_extensions:
                raise ValueError("Extension '%s' is not supported, use %s.",
                                 extension, ', '.join(supported_extensions))
        chain = self.chain.copy()

        #plot geweke has no meaning with chi_sorted
        if self.method is not 'chi_sorted':
            if fontsize is None:
                fontsize = plot.rcParams['font.size']

            fig = plot.figure(2, figsize=(16, 9))

            names = chain.dtype.names
            short_dict = self.galaxy.short_dict()

            plot.clf()  # clear current figure
            plot.subplots_adjust(wspace=0.32, hspace=0.32,
                                 bottom=0.05, top=0.95, left=0.05, right=0.95)
            n = np.size(names)
            score = np.zeros( len(self.galaxy.names), dtype='<f8')

            rows = 3
            cols = int(math.ceil((n-1) / (rows - 1.)))

            for i, par in enumerate(self.galaxy.names):
                #if i<n-1:
                # Parameters
                plot.subplot2grid((rows, cols), (int(math.floor(i / cols)), i % cols))
                if i==0:
                    plot.ylabel(r'$\sigma$')

                if full_chain is False:
                    zscore = self._geweke_score(self.sub_chain[par], intervals=Nintervals)
                else:
                    zscore = self._geweke_score(self.chain[par], intervals=Nintervals)
                _x = zscore[:,0]
                _y = zscore[:,1]
                if full_chain is False:
                    plot.plot(_x, _y, color='k', label=r'Sub-chain')
                else:
                    plot.plot(_x, _y, color='k', label=r'Full-chain')

                plot.plot(_x[0], _y[0], 'ks',ms=4,label=r'  Initial at 100\%')
                plot.plot(_x[-1],_y[-1],'ko',ms=7,label=r'  Final at $>50$\%')

                #in red masked array
                _mask = np.ma.masked_inside(_y, -Nsigma, Nsigma)
                score[i] = _mask.mask.sum()/Nintervals #should be 1 if True
                if score[i]==1.0:
                    self.logger.info(r'Parameter {:s} has converged ? {:.2f}'.format(par, score[i]))
                else:
                    self.logger.warning(r'Parameter {:s} has not converged ? {:.2f}'.format(par, score[i]))

                plot.plot(_x,_mask, color='r')
                plot.plot(_x[-1],_mask[-1],'ro',ms=7)

                plot.ylim([-3.5,3.5])
                plot.axhline(-Nsigma,ls='-.',lw=2.5)
                plot.axhline(+Nsigma,ls='-.',lw=2.5, label=r'%s$\sigma$' % (Nsigma) )
                plot.axhline(-1,ls='-.',lw=1)
                plot.axhline(+1,ls='-.',lw=1,label=r'$1\sigma$')

                title = r"%s" % (short_dict[par])
                title += "(%s)" % (self.galaxy.unit_dict()[par])
                plot.title(title, fontsize=fontsize)
            #elif i==n-1:
            plot.xlabel('Start index')
            plot.legend(loc='lower center', fontsize=fontsize, bbox_to_anchor=(0.0, -0.8), ncol=2, fancybox=True, shadow=True)

            self.convergence=Table(score,names=self.galaxy.names)
            #fig.subplots_adjust(wspace=0.3)

            if filepath is None:
                plot.show()
            else:
                plot.savefig(filepath)
                plot.close()
        else:
            self.convergence = None
            self.logger.warning('Plot geweke not allowed')


    def plot_correlations(self):
        """
        old routine
        :return:
        """
        self.logger.warning('plot_deprecated() \n use plot_corner() instead (requires corner package)')
        return None

    def plot_corner(self, filepath=None, smooth=None, nsigma=4, fontsize=10):
        '''
        using corner package to plot chain
        filepath: string [default:None]
            filename for saving the plot
        smooth: [default: None]
            smooth option from corner
        nsigma: int [default:4]
            N-sigma (nsigma x stdev) to determine the ranges for each plot
            using the range option from corner

        '''

        if np.size(self.chain)>200:
            try:
                import corner
                corner_true = True
            except:
                corner_true = False
                self.logger.warning('plot_corner(): Need corner package for this')

            if corner_true:

                fig = plot.figure(3)
                plot.plot()
                fig.set_figwidth(10)
                fig.set_figheight(10)


                shorts_dict = self.galaxy.short_dict()
                shorts_dict.pop('x') #remove x
                shorts_dict.pop('y') #remove y
                shorts_dict.pop('z') #remove z
                params = shorts_dict.keys()

                show = list(params)
                s=show[0]
                ls=[shorts_dict[s]] # also

                #n sigma
                r = [(self.galaxy[s] - self.galaxy.stdev[s] * nsigma, \
                      self.galaxy[s] + self.galaxy.stdev[s] * nsigma)]
                best = list([self.galaxy[s]])
                for s in show[1:]:
                    r.append((self.galaxy[s] - self.galaxy.stdev[s] * nsigma, \
                      self.galaxy[s] + self.galaxy.stdev[s] * nsigma))
                    best.append(self.galaxy[s])
                    ls.append(shorts_dict[s])

                if fontsize is None:
                    fontsize = plot.rcParams['font.size']

                #@fixme how to make onto fig=plot.figure(3)
                corner.corner(self.sub_chain[show].as_array().tolist(),\
                              labels=ls, truths=best, range=r,  \
                              smooth=smooth, label_kwargs={'fontname':'sans-serif', 'fontsize': fontsize})

                if filepath is None:
                    plot.show()
                else:
                    plot.savefig(filepath)
                    plot.close()

            else:
                self.logger.warning('package corner not installed; plot will be skipped')

        else:
            corner_true = False
            self.logger.warning('plot_corner skipped, chain too small (<200)')

        return corner_true


    def plot_true_vfield(self, filepath=None, mask=None, contours=None,
                         fontsize=10, slitwidth=3):
        """
        plot the 2d maps and the 1d along the major axis

        mask: [optional] 2d nd-array
        mask for display purposes
            None: [default] default mask is flux>max(flux)/20.
            1: not apply any mask
            nd array: mask to be used

        contours: [optional] 2d nd-array
        external map to overlay

        fontsize: int
        to change the fontsize

        slitwidth: 3 [default] the slit width used to extract 1d profile

        """

        fmap = self.true_flux_map.data
        vmap = self.true_velocity_map.data
        smap = self.true_disp_map.data
        try:
            vmax = self.galaxy.maximum_velocity
        except:
            try:
                vmax = self.galaxy.virial_velocity
            except:
                try:
                    vmax = self.galaxy.halo_velocity
                except:
                    vmax = None
                    self.logger.warning("Galaxy has no Vmax Vvir halo_velocity")

        if mask is None:
            # default mask is flux>max(flux)/30.
            mask = (fmap > bn.nanmax(fmap) / 50.) * (fmap != 0)

        xc = self.galaxy.x
        yc = self.galaxy.y

        pixscale = self.instrument.xy_step

        ny, nx = np.shape(fmap)

        #matplotlib display tickmarks on half pixels
        x0 = pixscale * (-xc -0.5)
        x1 = pixscale * (nx - xc -0.5)
        y0 = pixscale * (-yc -0.5)
        y1 = pixscale * (ny - yc -0.5)
        extent = [x0, x1, y0, y1]

        if fontsize is None:
            fontsize = plot.rcParams['font.size']

        plot.figure(1,figsize=(16,9))
        plot.clf()

        plot.subplot(2, 3, 1)
        ax1=self._plot2dimage(fmap / mask,
                          vmin=0, vmax=np.nanmax(fmap),
                          xlabel=r'$\Delta \alpha$(")',
                          ylabel=r'$\Delta \delta$(")', interpolation='nearest',
                          extent=extent, contour=contours, title='True Flux map')

        plot.subplot(2, 3, 2)
        ax2=self._plot2dimage(vmap / mask,
                          vmin=np.nanmin(vmap), vmax=bn.nanmax(vmap),
                          xlabel=r'$\Delta \alpha$(")',
                          ylabel=r'$\Delta \delta$(")', interpolation='nearest',
                          extent=extent, contour=contours, title='True Vel. map')

        plot.subplot(2, 3, 3)
        s0 = bn.nanmin(smap)
        s1 = bn.nanmax(smap)*1.1
        ax3=self._plot2dimage(smap / mask, vmin=s0, vmax=s1,
                          xlabel=r'$\Delta \alpha$(")',
                          ylabel=r'$\Delta \delta$(")',
                          extent=extent, contour=contours,
                          interpolation='nearest', title='True Disp. map')

        xx, slice_f = self._slit(fmap, slitwidth, ax=ax1)
        good = (slice_f != 0)

        plot.subplot(2, 3, 4)
        if self.error_maps:
            fmap_err = self.true_flux_map_error.data
            ax4=self._plot2dimage(fmap/fmap_err/mask,
                              vmin=0, vmax=np.nanmax(fmap/fmap_err),
                              xlabel=r'$\Delta \alpha$(")',
                              ylabel=r'$\Delta \delta$(")',
                              extent=extent, contour=fmap,
                              interpolation='nearest', title='True Fmap SNR')
        else:
            plot.plot(pixscale * xx[good], np.log10(slice_f[good]), 'k-')
            plot.xlim([-3, 3])
            plot.ylim([np.min(np.log10(slice_f[good])), bn.nanmax(np.log10(1.3 * slice_f[good]))])
            plot.xlabel(r'$\delta x$')
            plot.ylabel(r'$\log$ I(r)')

        plot.subplot(2, 3, 5)
        xx, slice_v = self._slit(vmap, slitwidth, ax=ax2)

        if self.error_maps:
            vmap_err = self.true_velocity_map_error.data
            ax5=self._plot2dimage(abs(vmap/vmap_err)/mask,
                            vmin=0,
                            vmax=np.nanmax(abs(vmap/vmap_err)),
                              xlabel=r'$\Delta \alpha$(")',
                              extent=extent, contour=fmap,
                              interpolation='nearest', title='True Vmap SNR')
        else:
            plot.plot(pixscale * xx[good], slice_v[good], 'k-')
            plot.xlim([-3, 3])
            if vmax is not None:
                plot.ylim([-vmax, vmax])
                plot.axhline(vmax * np.sin(np.radians(self.galaxy.inclination)), color='k', ls='--')
                plot.axhline(-vmax * np.sin(np.radians(self.galaxy.inclination)), color='k', ls='--')
            plot.xlabel(r'$\delta x$')
            plot.ylabel(r'$V_{los}(x)$')

        plot.subplot(2, 3, 6)
        xx, slice_s = self._slit(smap, slitwidth, ax=ax3)

        if self.error_maps:
            smap_err = self.true_disp_map_error.data
            ax6=self._plot2dimage(smap / smap_err / mask,
                              vmin=0,
                              vmax=np.nanmax(smap / smap_err),
                              xlabel=r'$\Delta \alpha$(")',
                              extent=extent, contour=fmap,
                              interpolation='nearest', title='True Disp SNR')
        else:
            plot.plot(pixscale * xx[good], slice_s[good], 'k-')
            if 'velocity_dispersion' in self.galaxy.names:
                plot.axhline(self.galaxy.velocity_dispersion, ls='-.', label='Vturb')
            plot.xlim([-3, 3])
            plot.ylim([0, s1])
            plot.xlabel(r'$\delta x$')
            plot.legend()

        plot.tight_layout()

        if filepath is None:
            plot.show()
        else:
            plot.savefig(filepath)
            plot.close()

            rotation_curve = [pixscale * xx[good], xx[good]/self.galaxy.radius, slice_f[good], slice_v[good], slice_s[good]]
            rotation_name=re.sub('true_maps','true_Vrot',filepath[:-4])+'.dat'
            asciitable.write(rotation_curve, \
                             output=rotation_name,Writer=asciitable.FixedWidth, \
                             names=['dx_arcsec', 'rad_Re', 'flux_slit', 'v_kms', 'sig_kms'], overwrite=True)


    def plot_obs_vfield(self, filepath=None, mask=None, contours=None,
                         fontsize=10, slitwidth=3):
        """
        plot the observed 2d maps and the 1d along the major axis

        mask: [optional] 2d nd-array
        mask for display purposes
            None: [default] default mask is flux>max(flux)/20.
            1: not apply any mask
            nd array: mask to be used

        contours: [optional] 2d nd-array
        external map to overlay

        fontsize: int
        to change the fontsize

        """
        #Fmap, Vmap, Smap = self._make_moment_maps(self.convolved_cube, mask=True)
        Fmap = self.obs_flux_map
        Vmap = self.obs_velocity_map
        Smap = self.obs_disp_map
        fmap = Fmap.data
        vmap = Vmap.data
        smap = Smap.data

        try:
            vmax = self.galaxy.maximum_velocity
        except:
            try:
                vmax = self.galaxy.virial_velocity
            except:
                try:
                    vmax = self.galaxy.halo_velocity
                except:
                    vmax = None
                    self.logger.warning("Galaxy has no Vmax Vvir halo_velocity")

        if mask is None:
            # default mask is flux>max(flux)/30.
            mask = (fmap > bn.nanmax(fmap) / 50.) * (fmap != 0)

        xc = self.galaxy.x
        yc = self.galaxy.y

        pixscale = self.instrument.xy_step

        ## Matplotlib uses a grid with tickmarks on the middle of the pixels
        ## hence the -0.5
        ny, nx = np.shape(fmap)

        x0 = pixscale * (-xc -0.5)
        x1 = pixscale * (nx - xc -0.5)
        y0 = pixscale * (-yc -0.5)
        y1 = pixscale * (ny - yc -0.5)
        extent = [x0, x1, y0, y1]

        if fontsize is None:
            fontsize = plot.rcParams['font.size']

        plot.figure(1, figsize=(16,9))
        plot.clf()
        plot.title("EXPERIMENTAL products; not used by 3DMODEL ", color='red', fontsize=fontsize)

        plot.subplot(2, 3, 1)
        ax1=self._plot2dimage(fmap / mask,
                          vmin=0, vmax=np.nanmax(fmap),
                          xlabel=r'$\Delta \alpha$(")',
                          ylabel=r'$\Delta \delta$(")', interpolation='nearest',
                          extent=extent, contour=contours, title='Obs. Flux map')

        plot.subplot(2, 3, 2)
        ax2=self._plot2dimage(vmap / mask,
                          vmin=np.nanmin(vmap), vmax=bn.nanmax(vmap),
                          xlabel=r'$\Delta \alpha$(")',
                          ylabel=r'$\Delta \delta$(")', interpolation='nearest',
                          extent=extent, contour=contours, title='Obs. Vel. map')

        plot.subplot(2, 3, 3)
        s1 = bn.nanmax(smap)*1.1
        try:
            s0 = bn.nanmin(smap)  #- 2 * bn.nanstd(smap)
        except:
            s0 = np.nanmin(smap)  #- 2 * np.nanstd(smap)
        ax3=self._plot2dimage(smap / mask, vmin=s0, vmax=s1,
                          xlabel=r'$\Delta \alpha$(")',
                          ylabel=r'$\Delta \delta$(")',
                          extent=extent, contour=contours,
                          interpolation='nearest', title='Obs. Disp. map')

        plot.subplot(2, 3, 4)
        xx, slice_v = self._slit(fmap, slitwidth, ax=ax1)
        good = (slice_v != 0)
        plot.plot(pixscale * xx[good], np.log10(slice_v[good]), 'k-')
        plot.xlim([-3, 3])
        plot.ylim([np.min(np.log10(slice_v[good])), bn.nanmax(np.log10(1.3 * slice_v[good]))])
        plot.xlabel(r'$\delta x$')
        plot.ylabel(r'$\log$ I(r)')

        plot.subplot(2, 3, 5)
        xx, slice_v = self._slit(vmap, slitwidth, ax=ax2)
        plot.plot(pixscale * xx[good], slice_v[good], 'k-')
        plot.xlim([-3, 3])
        if vmax is not None:
            plot.ylim([-vmax, vmax])
            plot.axhline(vmax * np.sin(np.radians(self.galaxy.inclination)), color='k', ls='--')
            plot.axhline(-vmax * np.sin(np.radians(self.galaxy.inclination)), color='k', ls='--')
        plot.xlabel(r'$\delta x$')
        #p.ylabel(r'$V_{los}(x)$')

        plot.subplot(2, 3, 6)
        xx, slice_v = self._slit(smap, slitwidth, ax=ax3)
        plot.plot(pixscale * xx[good], slice_v[good], 'k-')
        if 'velocity_dispersion' in self.galaxy.names:
            plot.axhline(self.galaxy.velocity_dispersion, ls='-.', label='Vturb')
        plot.xlim([-3, 3])
        plot.ylim([0, s1])
        plot.xlabel(r'$\delta x$')
        plot.legend()

        plot.tight_layout()

        if filepath is None:
            plot.show()
        else:
            plot.savefig(filepath)
            plot.close()

    def plot_images(self, filepath=None, z_crop=None):
        """
        Plot a mosaic of images of the cropped (along z) cubes,
        and then either show it or save it to a file.

        filepath: string
            If specified, will write the plot to a file instead of showing it.
            The file will be created at the provided absolute or relative filepath.
            The extension of the file must be either png or pdf.
        z_crop: None|int
            The maximum and total length of the crop (in pixels) along z,
            centered on the galaxy's z position.
            If you provide zero or an even value (2n),
            the closest bigger odd value will be used (2n+1).
            By default, will not crop.
        """

        if self.chain is None:
            raise RuntimeError(self.NO_CHAIN_ERROR)

        if filepath is not None:
            name, extension = os.path.splitext(filepath)
            supported_extensions = ['.png', '.pdf']
            if not extension in supported_extensions:
                raise ValueError("Extension '%s' is not supported, "
                                 "you may use one of %s",
                                 extension, ', '.join(supported_extensions))

        self._plot_images(self.cube, self.galaxy, self.convolved_cube,
                          self.deconvolved_cube, self.residuals_cube,
                          z_crop=z_crop)

        if filepath is None:
            plot.show()
        else:
            plot.savefig(filepath)
            plot.close()

    def _plot_images(self, cube, galaxy, convolved_cube, deconvolved_cube,
                     residuals_cube, z_crop=None):
        """
        Plot a mosaic of images of the cropped (along z) cubes.

        z_crop: None|int
            The maximum and total length of the crop (in pixels) along z,
            centered on the galaxy's z position.
            If you provide zero or an even value (2n),
            the closest bigger odd value will be used (2n+1).
            By default, will not crop.
        """
        if z_crop is None:
            zmin = 0
            zmax = cube.shape[0] - 1
        else:
            if not (z_crop & 1):
                z_crop += 1
            z0 = galaxy.z
            zd = (z_crop - 1) / 2
            zmin = max(0, z0 - zd)
            zmax = z0 + zd + 1

        fig = plot.figure(1, figsize=(16, 9))
        plot.clf()
        plot.subplots_adjust(wspace=0.25, hspace=0.25, bottom=0.05, top=0.95, left=0.05, right=0.95)

        # MEASURE
        sub = fig.add_subplot(2, 2, 1)
        sub.set_title('Measured')
        measured_cube_cropped = cube.data[zmin:zmax, :, :]
        image_measure = (measured_cube_cropped.sum(0) / measured_cube_cropped.shape[0])
        plot.imshow(image_measure, interpolation='nearest', origin='lower')
        plot.xticks(fontsize=8)
        plot.yticks(fontsize=8)
        colorbar = plot.colorbar()
        colorbar.ax.tick_params(labelsize=8)

        # CONVOLVED
        sub = fig.add_subplot(2, 2, 2)
        sub.set_title('Convolved')
        convolved_cube_cropped = convolved_cube.data[zmin:zmax, :, :]
        image_convolved = (convolved_cube_cropped.sum(0) / convolved_cube_cropped.shape[0])
        plot.imshow(image_convolved, interpolation='nearest', origin='lower')
        plot.xticks(fontsize=8)
        plot.yticks(fontsize=8)
        colorbar = plot.colorbar()
        colorbar.ax.tick_params(labelsize=8)

        # DECONVOLVED
        sub = fig.add_subplot(2, 2, 3)
        sub.set_title('Deconvolved')
        deconvolved_cube_cropped = deconvolved_cube.data[zmin:zmax, :, :]
        image_deconvolved = (deconvolved_cube_cropped.sum(0) / deconvolved_cube_cropped.shape[0])
        plot.imshow(image_deconvolved, interpolation='nearest', origin='lower')
        plot.xticks(fontsize=8)
        plot.yticks(fontsize=8)
        colorbar = plot.colorbar()
        colorbar.ax.tick_params(labelsize=8)

        # ERROR
        sub = fig.add_subplot(2, 2, 4)
        sub.set_title('Error')
        square_error_cube_cropped = residuals_cube.data[zmin:zmax, :, :]
        nz = square_error_cube_cropped.shape[0]
        #normalized error image in sigmas:
        image_error = (square_error_cube_cropped.sum(0) / nz) * np.sqrt(nz)
        #vmin = np.amin(image_measure)
        #vmax = np.amax(image_measure)
        #plot.imshow(image_error, interpolation='nearest', origin='lower', vmin=vmin, vmax=vmax)
        plot.imshow(image_error, vmin=-2.5, vmax=2.5, interpolation='nearest', origin='lower')
        plot.xticks(fontsize=8)
        plot.yticks(fontsize=8)
        colorbar = plot.colorbar()
        colorbar.ax.tick_params(labelsize=8)

        return fig

    def _plot2dimage(self, image, vmin=-10, vmax=10, pos=None,
                    xlabel=None, ylabel=None, contour=None,
                    extent=None, title=None, interpolation=None, **kwargs):
        """
          Plot a 2D image with colorbars.
        """

        # Matplotlib's default origin is not ours, we need to use 'lower'
        origin = 'lower'

        if extent is not None:
            plot.imshow(image, aspect='equal', vmin=vmin, vmax=vmax, cmap=plot.cm.jet,
                        extent=extent, origin=origin,
                        interpolation=interpolation)
            plot.axhline(0,color='k',lw=1,alpha=0.5)
            plot.axvline(0,color='k',lw=1,alpha=0.5)
            if contour is not None:
                cmax = bn.nanmax(contour)
                plot.contour(contour, extent=extent, color='k',
                             levels=[cmax / 10., cmax / 5., cmax / 2.],
                             origin=origin)
        else:
            plot.imshow(image, aspect='equal', vmin=vmin, vmax=vmax, cmap=plot.cm.jet,
                        origin=origin)
            plot.axhline(0,color='k',lw=1,alpha=0.5)
            plot.axvline(0,color='k',lw=1,alpha=0.5)
            if contour is not None:
                cmax = bn.nanmax(contour)
                plot.contour(contour, levels=[cmax / 10., cmax / 5., cmax / 2.],
                             color='k', origin=origin)
        if title is not None:
            plot.title(title)
        ax = plot.gca()
        v = ax.axis()
        if pos is not None:
            plot.plot(pos[0], pos[1], 'k+', ms=25, mew=3)
        plot.axis(v)

        if xlabel is not None:
            plot.xlabel(xlabel)
        if ylabel is not None:
            plot.ylabel(ylabel)
            #ax=p.gca()

        # Add colorbar
        cax = inl.inset_axes(
            ax,
            width="80%",  # width = 10% of parent_bbox width
            height="10%",  # height : 50%
            #borderpad=1,
            #    bbox_to_anchor=(1,0,1,2),
            # bbox_transform=ax.transAxes,
            loc=1)
        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
        if vmax > 1:
            tick_range = np.rint(np.r_[vmin:vmax:5j])
        else:
            tick_range = (np.r_[vmin:vmax:5j])

        # matplotlib will crash and burn if ticks are not unique
        tick_range = np.unique(tick_range)

        cb = mpl.colorbar.ColorbarBase(cax, cmap=plot.cm.jet,
                                       norm=norm, orientation='horizontal',
                                       ticks=tick_range[1:-1])

        #cax.xaxis.set_major_locator( MaxNLocator(nbins = 6) )
        return ax

    def _slit(self, mapdata, slit_width, reshape=False, ax=None):
        """
        Create 1d profile along the major axis.
        Note that PA = 0 is vertical, anti-clockwise from y-axis.
        """

        #orig code
        #ang=np.radians(PA)
        #prof=np.where(s<=(slit_width/2.0+0.5),data,0.)
        #profile=ndimage.rotate(prof,PA,mode='constant',order=1).sum(1)
        #profile=profile/np.max(profile)

        y, x = np.indices(mapdata.shape)
        ang = -np.radians(self.galaxy.pa - 90)

        dx = x - self.galaxy.x
        dy = y - self.galaxy.y
        dx_p = dx * np.cos(ang) - dy * np.sin(ang)
        dy_p = dx * np.sin(ang) + dy * np.cos(ang)

        if (ax!=None):
            #plot slit
            #axis labels in arcsec
            xlim=ax.get_xlim()
            ylim=ax.get_ylim()
            pixscale = self.instrument.xy_step
            xar= (x[0] - self.galaxy.x) * pixscale
            ang = np.radians(self.galaxy.pa + 90.)
            xx= xar * np.cos(ang) - (slit_width/2.) * pixscale * np.sin(ang)
            yy= xar * np.sin(ang) + (slit_width/2.) * pixscale * np.cos(ang)
            ax.plot(xx,yy,'-',lw=2,c='grey',alpha=0.6)
            xx= xar * np.cos(ang) + (slit_width/2.) * pixscale * np.sin(ang)
            yy= xar * np.sin(ang) - (slit_width/2.) * pixscale * np.cos(ang)
            ax.plot(xx,yy,'-',lw=2,c='grey',alpha=0.6)
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

        slit_vert = np.abs(dy_p)  # used for PA=0
        slit_horiz = dx_p  # distance along the slit

        slit_data     =  np.where(slit_vert <= (slit_width / 2.0 ), mapdata, 0.)
        slit_distance =  np.where(slit_vert <= (slit_width / 2.0 ), slit_horiz,0.)
        # PA is anti-clockwise, rotate is clockwise
        rot_ang = self.galaxy.pa + 90  # to have horizontal slit image
        slit_rot = ndimage.rotate(slit_data, rot_ang, mode='constant', order=0,reshape=reshape)  # 0 because spline introduces errors!
        dist_rot = ndimage.rotate(slit_distance, rot_ang, mode='constant', order=0,reshape=reshape)


        a =  np.sum(slit_rot != 0,axis=(0)) #nansum returns a bool with bottelneck; not with numpy!!
        profile = np.where(a!=0, bn.nansum(slit_rot, axis=(0)) / a, 0.)  #mean of the slit

        b =  np.sum(dist_rot != 0,axis=(0))  # pixels not zero and not nan
        xaxis =  np.where(b!=0, bn.nansum(dist_rot, axis=(0)) / b, 0.)

        return xaxis, profile

    def _make_maps_Epinat(self, cube=None, mask=False, cut_level=50, use_conv='astropy'):
        """
        make observed maps from true maps using formulas in Epinat 2010

        :param cube: input cube for computing masking
        :param mask: boolean [default: None] to apply masking
        :return:
        """

        if (cube is None):
                cube = self.convolved_cube

        if mask is True:
            # default mask is flux>max(flux)/30.
            mask = (bn.nansum(cube.data,0) > bn.nanmax(cube.data.sum(0)) / cut_level )
        else:
            mask = np.ones_like(cube.data[1,:,:])

        #Eq A15 Epinat
        psf2d = self.instrument.psf3d.sum(axis=0)
        if use_conv is 'scipy' :
            logger.debug("Using scipy %s for convolution" % (scipy.__version__))
            if scipy.__version__>'1.3.0':
                Fmap = conv2d_scipy(self.true_flux_map.data, psf2d, mode='same', method='auto')
            else:
                Fmap = conv2d_scipy(self.true_flux_map.data, psf2d, mode='same')
        elif use_conv is 'astropy':
            logger.debug("Using astropy %s for convolution" % (astropy.__version__))
            Fmap = conv2d_astro(self.true_flux_map.data, psf2d)
        else:
            Fmap, psf2d_fft = conv2d_intern(self.true_flux_map.data, psf2d, compute_fourier=True)

        #Eq A16 Epinat
        V1 = self.true_velocity_map.data * self.true_flux_map.data          # V.M
        if use_conv is 'scipy':
            if scipy.__version__>'1.3.0':
                Vtemp = conv2d_scipy(V1, psf2d, mode='same', method='auto')
            else:
                Vtemp = conv2d_scipy(V1, psf2d, mode='same')
        elif use_conv is 'astropy':
            Vtemp = conv2d_astro(V1, psf2d)
        else:
            Vtemp, psf2d_fft = conv2d_intern(V1, psf2d, compute_fourier=True)
        Vmap = Vtemp / Fmap

        #Eq A17 Epinat
        S1 = self.true_disp_map.data**2 * self.true_flux_map.data
        if use_conv is 'scipy':
            if scipy.__version__>'1.3.0':
                Stemp = conv2d_scipy(S1, psf2d, mode='same', method='auto')
            else:
                Stemp = conv2d_scipy(S1, psf2d, mode='same')
        elif use_conv is 'astropy':
            Stemp = conv2d_astro(S1, psf2d)
        else:
            Stemp, psf2d_fft = conv2d_intern(S1, psf2d, compute_fourier=True)

        V2 = self.true_velocity_map.data**2 * self.true_flux_map.data
        if use_conv is 'scipy':
            if scipy.__version__>'1.3.0':
                Vtemp = conv2d_scipy(V2, psf2d, mode='same', method='auto')
            else:
                Vtemp = conv2d_scipy(V2, psf2d, mode='same')
        elif use_conv is 'astropy':
            Vtemp = conv2d_astro(V2, psf2d)
        else:
            Vtemp, psf2d_fft = conv2d_intern(V2, psf2d, compute_fourier=True)

        Smap_sq = Stemp / Fmap + Vtemp / Fmap - Vmap**2
        Smap = np.sqrt(Smap_sq)

        self.obs_flux_map = HyperCube(Fmap * mask)
        self.obs_velocity_map = HyperCube(Vmap * mask)
        self.obs_disp_map = HyperCube(Smap * mask)

        return self.obs_flux_map, self.obs_velocity_map, self.obs_disp_map

    def _make_moment_maps(self, cube=None, mask=False, cut_level=50, parameters=None, instrument=None, remove_LSF_from_disp=False):
        """
        make moment maps from a noiseless cube
        assumes no continuum to be removed
        :param mask: boolean [default: None] to apply masking
        :param cube: HyperspectralCube
        :return:
        """

        if (cube is None):
                cube = self.convolved_cube

        if (instrument is None):
            vstep = self.instrument.z_step_kms
            lsf_vector = self.instrument.lsf.as_vector(cube)
            if remove_LSF_from_disp:
                lsf_fwhm = self.instrument.lsf.fwhm / self.instrument.z_step #in pixel  !
        else:
            vstep = instrument.z_step_kms
            lsf_vector = instrument.lsf.as_vector(cube)
            if remove_LSF_from_disp:
                x=np.arange(np.size(lsf_vector))
                sigma = np.sqrt(np.sum(x**2*lsf_vector)-np.sum(x*lsf_vector)**2) #in pixel !
                lsf_fwhm = sigma * 2.35

        zgrid, _, _ = np.indices(cube.shape)

        if parameters is None:
                zo = self.galaxy.z
        else:
                zo = parameters.z

        if mask is True:
            # default mask is flux>max(flux)/30.
            mask = (cube.data > bn.nanmax(cube.data) / cut_level )
        else:
            mask = np.ones_like(cube.data)

        #if the line is a doublet use a weighted average
        if self.model.line is not None:
                line = self.model.line
                delta = 3e5 * (line['wave'][1]-line['wave'][0])/(line['wave'][0]+line['wave'][1])*2
                r1 = line['ratio'][0]
                r2 = line['ratio'][1]
                p = r1 / (r1+r2) #weight of first/blue line
                # weighted averaged zo
                #zo_avg =  ((zo-delta/vstep) * r1 + zo * r2) / (r1 + r2)
                #print zo,zo-delta/vstep,zo_avg
                #zo = zo_avg

                #in km/s
                # M1 = p * mu1 + (1-p) * mu2 ; m1 = mu2-Delta
                # mu2 = (M1 - p mu1 ) / (1-p)
                # mu2 = M1 + p delta
                Fmap = bn.nansum(cube.data * mask,axis=0)
                cube_norm = np.where(np.isfinite(cube.data / Fmap), cube.data / Fmap, 0)
                M1map = bn.nansum(cube_norm * mask * (zgrid - zo), axis=0)   * vstep
                Vmap = (M1map + p * delta) # Vmap

                # M2 = p * (mu1^2+sig^2) + (1-p) * (mu2^2+sig^2); m1 = mu2-Delta
                # M2 = sig^2 + (p mu1^2 + (1-p) mu2^2)
                # M2 - p (mu2-delta)**2 - (1-p) mu2^2) = sig^2
                M2map = bn.nansum(cube.data / Fmap * mask * (zgrid -zo)**2, axis=0)  * vstep**2
                S2map = M2map - (Vmap-delta)**2 * p - (Vmap)**2 * (1-p)
                Smap = np.sqrt(S2map)

        else:
                Fmap = bn.nansum(cube.data * mask,axis=0)
                cube_norm = np.where(np.isfinite(cube.data / Fmap), cube.data / Fmap, 0)
                Vmap = bn.nansum(cube_norm * mask * (zgrid-zo), axis=0)  * vstep     # mu
                S2map = bn.nansum(cube_norm * mask *(zgrid-zo)**2, axis=0)  * vstep**2 # mu^2 + sigma^2
                Smap = np.sqrt(S2map-Vmap**2)

        #Remove instrument LSF
        if (remove_LSF_from_disp):
            self.logger.info('removing LSF from dispersion map in quadrature; with FWHM %.5f pix' % (lsf_fwhm) )
            Smap=np.sqrt(Smap**2 - (lsf_fwhm/2.35 * vstep)**2)


        self.obs_flux_map = HyperCube(Fmap)
        self.obs_velocity_map = HyperCube(Vmap)
        self.obs_disp_map = HyperCube(Smap)

        return self.obs_flux_map, self.obs_velocity_map, self.obs_disp_map

    def _geweke_score(self, x, first=.1, last=.5, intervals=20):
        """
        Return z-scores for convergence diagnostics.
        Compare the mean of the first % of series with the mean of the last % of
        series. x is divided into a number of segments for which this difference is
        computed. If the series is converged, this score should oscillate between
        -1 and 1.
        Parameters
        ----------
        x : array-like
          The trace of some stochastic parameter.
        first : float
          The fraction of series at the beginning of the trace.
        last : float
          The fraction of series at the end to be compared with the section
          at the beginning.
        intervals : int
          The number of segments.
        maxlag : int
          Maximum autocorrelation lag for estimation of spectral variance
        Returns
        -------
        scores : list [[]]
          Return a list of [i, score], where i is the starting index for each
          interval and score the Geweke score on the interval.
        Notes
        -----
        The Geweke score on some series x is computed by:
          .. math:: \frac{E[x_s] - E[x_e]}{\sqrt{V[x_s] + V[x_e]}}
        where :math:`E` stands for the mean, :math:`V` the variance,
        :math:`x_s` a section at the start of the series and
        :math:`x_e` a section at the end of the series.
        References
        ----------
        Geweke (1992)
        """
        if np.ndim(x) > 1:
            return [self._geweke_score(y, first, last, intervals) for y in np.transpose(x)]

        # Filter out invalid intervals
        if first + last >= 1:
            raise ValueError(
                "Invalid intervals for Geweke convergence analysis",
                (first, last))

        # Initialize list of z-scores
        zscores = [None] * intervals

        # Starting points for calculations
        starts = np.linspace(0, int(len(x)*(1.-last)), intervals).astype(int)

        # Loop over start indices
        for i,s in enumerate(starts):

            # Size of remaining array
            x_trunc = x[s:]
            n = len(x_trunc)

            # Calculate slices
            first_slice = x_trunc[:int(first * n)]
            last_slice = x_trunc[int(last * n):]

            z = (first_slice.mean() - last_slice.mean())
            #to avoid numerical errors
            if np.var(last_slice)>1e-6*last_slice.mean()**2:
                z /= np.sqrt(np.var(first_slice) +
                         np.var(last_slice))
            else:
                z = 0
            zscores[i] = len(x) - n, z

        return np.array(zscores)

        #@fixme: unsupported
        #def film_images(self, name, frames_skipped=0, fps=25):
        #    #"""
        #This is still a WiP.

        #TODO:
        #    - print => logging
        #    - heavily document how to solve ffmpeg issue (even if it's not our job)

        #Generates an video of the evolution of plot_images() through the chain,
        #skipping [frames_skipped] between each draw, at [fps] frames per second.

        #.. warning::
        #    The generated file may take up a lot of disk space.

        #Known issue with ffmpeg :
        #https://stackoverflow.com/questions/17887117/python-matplotlib-basemap-animation-with-ffmpegwriter-stops-after-820-frames
        #Fix :
        #sudo apt-add-repository ppa:jon-severinsson/ffmpeg
        #sudo apt-get update
        #sudo apt-get install ffmpeg
        #"""
        #import matplotlib.animation as animation

        # Initial plot
        #fig = self._plot_images(self.cube, self.galaxy, self.convolved_cube,
        #                        self.deconvolved_cube, self.residuals_cube)

        ## Convert the chain to a list of parameters
        #chain_list = np.array(self.chain.tolist())
        #params = chain_list[:, :-1]  # remove the reduced chi (last element)

        ## Sanity check (otherwise, divisions by 0 will happen)
        #if len(params) < 2:
        #    raise RuntimeError("Chain is not long enough for video.")

        #self.logger.info("Encoding video...")

        #def animate(i, skipped, count, me, chain):
        #    """
        #    animation.FuncAnimation requires a function to generate the figure
        #    on each frame. This is the bottleneck of our film generation.

        #    i: int
        #        0 0 1 2 3 ... up to [count]
        #    skipped: int
        #        Between each frame, skip [skipped] parameters in the chain
        #    count: int
        #        The total [count] of frames to show.
        #    me: GalPaK3D
        #        Our local self.
        #    chain: ndarray
        #        The list of parameters, as long as the MCMC chain.
        #    """
        #    # Get the parameters from the chain
        #    index = int(math.floor(i / (1 + skipped)))
        #    galaxy = ModelParameters.from_ndarray(chain[index])
        #    # Print the progression in %
        #    sys.stdout.write("\r%d: %2.2f%%" % (i, 100. * i / (count - 1)))
        #    sys.stdout.flush()
        #    # Compute the cubes
        #    deconvolved_cube = me.create_clean_cube(galaxy, me.cube.shape)
        #    convolved_cube = me.instrument.convolve(deconvolved_cube.copy())
        #    residuals_cube = (me.cube - convolved_cube) / me.error_cube
        #    # Plot and return the plot
        #    # (this is the expensive step, and it can be optimized further)
        #    return me._plot_images(me.cube, galaxy, convolved_cube,
        #                          deconvolved_cube, residuals_cube)

        #if frames_skipped < 0:
        #    frames_skipped *= -1
        #frames_count = int(math.floor(len(params) / (1. + frames_skipped)))
        #ani = animation.FuncAnimation(fig, animate, frames_count,
        #                              fargs=(int(frames_skipped), frames_count,
        #                                     self, params),
        #                              repeat=False)

        ## Still unsure if that metadata is really written in the file
        #metadata = {
        #    'title': 'GalPaK\'s cubes timelapse',
        #    'author': 'galpak',
        #}
        #writer = animation.FFMpegWriter(fps=fps)
        #ani.save(name + '_images.avi', writer=writer, metadata=metadata)



import unittest
import numpy as np
from numpy import ndarray
import os, sys
sys.path.append('../')

import galpak
from tests.tools.numeric_assertions import NumericAssertions

GalaxyParameters = galpak.DefaultModel().parameters_class()

class GalaxyParametersTest(unittest.TestCase, NumericAssertions):

    longMessage = True

    def test_init_from_array(self):
        g = GalaxyParameters()
        g.from_ndarray(np.arange(len(g)))
        self.assertIsInstance(g, GalaxyParameters)

    def test_init_without_data(self):
        g = GalaxyParameters()
        self.assertIsNan(g.x)

    def test_init_with_data(self):
        g = GalaxyParameters(y=2.0)
        self.assertIsNan(g.x)
        self.assertEqual(g.y, 2.0)

    def test_init_from_ndarray(self):
        a = ndarray(shape=10, dtype=float)
        a[1] = 2.0
        g = GalaxyParameters()
        t = g.from_ndarray(a)
        # self.assertIsNan(g.x)
        # => instead yields 1.88532209e-316 or equalish infinitesimal value
        self.assertEqual(t.y, 2.0)

    def test_init_readwrite_file(self):

        a = ndarray(shape=10, dtype=float)
        a[1] = 2.0
        g = GalaxyParameters()
        g.from_ndarray(a)
        g.save_to_file('test_param.txt', overwrite=True)
        self.assertTrue(os.path.isfile('test_param.txt'))
        empty = GalaxyParameters()
        empty.from_file('test_param.txt')
        test=np.allclose(g,empty,equal_nan=True)
        self.assertTrue(test)
        os.system('rm -f test_param.txt')


    def test_accessors_and_mutators(self):

        # Explicit instantiation
        g1 = GalaxyParameters(0., 1., 2., 3., 4., 5., 6., 7., 8., 9.)

        # Testing accessors after instantiation
        self.assertEqual(0., g1[0])
        self.assertEqual(0., g1.x)
        self.assertEqual(0., g1['x'])
        self.assertEqual(1., g1[1])
        self.assertEqual(1., g1.y)

        # Mutating using numeric keys
        g1[0] = 11.
        self.assertEqual(11., g1[0], "Mutating a key-based value should change the key-based value")
        self.assertEqual(11., g1.x,  "Mutating a key-based value should change the attribute")

        # Mutating using attribute
        g1.x = 13.
        self.assertEqual(13., g1.x,  "Mutating an attribute should change the attribute")
        self.assertEqual(13., g1[0], "Mutating an attribute should change the key-based value")

        # Mutating using dict
        g1['x'] = 1.618
        self.assertEqual(1.618, g1.x,  "Mutating through dict should change the attribute")
        self.assertEqual(1.618, g1[0], "Mutating through dict should change the key-based value")
        self.assertEqual(1.618, g1['x'], "Mutating through dict should change the dict-based value")

        # Using another instance
        g2 = GalaxyParameters(9., 8., 7., 6., 5., 4., 3., 2., 1., 0.)

        # Sanity checks
        self.assertEqual(9., g1[9])
        self.assertEqual(0., g2[9])

    def test_copy(self):

        g1 = GalaxyParameters(0., 1., 2., 3., 4., 5., 6., 7., 8., 9.)
        g2 = g1.copy()

        self.assertEqual(g1[0], g2[0], "Copy should copy key-based values")
        self.assertEqual(g1.x, g2.x,   "Copy should copy attributes too")

    def test_to_string(self):

        # No assertions, just print the galaxy so we can see
        g = GalaxyParameters(0., 1., 2., 3., 4., 5., 6., 7., 8., 9.)
        w = GalaxyParameters(x=0., z=1.)  # warnings flags
        e = GalaxyParameters(z=0.02)  # stdev margin

        print("Using print()")
        print(g)
        print("Using repr()")
        print(repr(g))
        print("Long info")
        print(g.long_info())
        print("Structured info")
        print(g.structured_info())
        print("Colored info")
        print(g.colored_info(w))

        g.error = e
        print("Detailed info w/ stdev")
        print(g.long_info())
        print("Detailed info w/ % stdev")
        print(g.long_info(True))



# coding=utf-8

from __future__ import division

import numpy as np, sys
from astropy.table import Table

from .galaxy_parameters import  GalaxyParameters
from .hyperspectral_cube import HyperspectralCube as HyperCube
from .math_utils import merge_where_nan, median_clip, safe_exp

class MCMC:
    """
        An extension of the galpak class with the galpak MH algorithm (myMCMC class)
        and method used for emcee and external algorithm

        the following methods are for external samplers such as emcee:
            - lnprob: returned by __call__ / the full log-probability function: Ln = Priors x L(p)
            - loglike: the log L(p)
            - logpriors: the log priors with  uniform priors.
        the following methods are for multinest:
            - pyloglike: the log L(p)
            - pycube: transforms unit cube to uniform priors.
        the following methods are for dynesty:
            - loglike: the log L(p)
            - ptforms: transforms unit cube to uniform priors.
        the following methods are for the internal galpak MH algorithm
            - myMCMC

    """
    CHISTAT_VALID = ['gaussian', 'Mighell', 'Cstat', 'Neyman', 'Pearson']

    SAMPLING_VALID = ['Cauchy', 'Normal', 'AdaptiveCauchy',\
                      'DE', 'Snooker', 'walkers', 'walkersCauchy']

    MCMC_VALID = ['galpak', 'emcee_MH', 'emcee_walkers', 'dynesty', 'multinest']

    def __call__(self, x):
        """
            call for instance ()
        used for emcee
        """
        return self.lnprob(x)

    #For EMCEE
    def lnprob(self, params):
        """
            the full log-probability function: Ln = Priors x L(p)
        """

        #convert ndarray to Parameter class {for emcee}
        if not isinstance(params, GalaxyParameters):
            if isinstance(params, np.ndarray):
                params = self.model.Parameters().from_ndarray(params)
            else:
                self.logger.error("Param is not ModelParameter nor a ndarray")

        lp = self.logpriors(params)
        if not np.isfinite(lp):
            return -np.inf
        return lp+self.loglike(params)

    def loglike(self, params):
        """
            The log-likelihood function.
            log L ~ log Pi_i [1/Vi^0.5 exp(-0.5 (xi-m)^/Vi]
            log L ~ -0.5 * sum_i Vi - 0.5 sum_i (xi-m)^2/Vi
        """
        #return -0.5*self.compute_chi(params)
        #adding a constant
        return -0.5*self.compute_chi(params)

    def logpriors(self, params):
        """
        Define  uniform prior.
        Make sure we're inside the preset boundaries

        :param params:
        :return:
        """
        if params is None:
            self.logger.error("Param is None " )

        in_boundaries = (params >= self.min_boundaries) * (params <= self.max_boundaries)
        if (in_boundaries==False).as_vector().any():
            return -np.inf
        else:
            self.model.sanitize_parameters(params)
            return 0.0
    #for PyMultinest
    def pyloglike(self, cube, ndim, nparams):
        """
        wrap for pymultinest
        """
        vect = [cube[i] for i in range(ndim)]
        params = self.model.Parameters().from_ndarray(vect)
        self.model.sanitize_parameters(params)

        return self.loglike(params)

    def pycube(self, cube, ndim, nparams):
        """
        Transforms samples `u` drawn from the unit cube to samples to those
        from our uniform prior

        for pymultinest
        """
        for i in range(ndim):
            cube[i] = (self.max_boundaries - self.min_boundaries)[i] * cube[i] + self.min_boundaries[i]

    #For Dynesty
    def ptform(self, u):
        """Transforms samples `u` drawn from the unit cube to samples to those
        from our uniform prior

        for Dynesty
        """
        return (self.max_boundaries - self.min_boundaries) * u + self.min_boundaries

    def compute_chi(self, params):
        # Sanitize the parameters
        #
        if not isinstance(params, GalaxyParameters):
            if isinstance(params, np.ndarray):
                params = self.model.Parameters().from_ndarray(params)

        self.model.sanitize_parameters(params) # to deal with circularity

        #dim_d = np.size(self.cube.data)
        ones = self.cube.data**2/self.variance_cube   #
        dim_d = np.isfinite(ones).sum() #count only data with finite values

        dim_p_free = dim_p = len(params)
        #already taken care
        #if self.model.flux_profile is not 'sersic':
        #    dim_p_free = dim_p -1
        if self.known_parameters is not None:
            fixed = np.where(np.isfinite(self.known_parameters),1,0).sum()
        else:
            _snr = self.galaxy.stdev / np.abs(self.galaxy)
            fixed = np.where(_snr<1e-6,1,0).sum()
        self.dim_p_free = dim_p_free - fixed

        self.Ndegree = (dim_d - self.dim_p_free -1)

        return self.chisq(self.chi_stat)(params)

    def chisq(self, chi_stat):

        # Chi2 satistics, default
        if (chi_stat == 'gaussian'):
            compute_chi_fct = lambda g: \
            np.nansum(self._compute_error_gauss(g)
            )
        # https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/XSappendixStatistics.html
        # Mighell http://adsabs.harvard.edu/abs/1999ApJ...518..380M
        # Humphrey 2009, http://adsabs.harvard.edu/abs/2009ApJ...693..822H
        # Mighell
        elif (chi_stat == 'Mighell'):
            compute_chi_fct = lambda g: \
            np.nansum(self._compute_error_Mighell(g)
            )
        # C-statistique
        elif (chi_stat == 'Cstat'):
            compute_chi_fct = lambda g: \
            np.nansum(self._compute_error_cstat(g)
            )
        # Neyman
        elif (chi_stat == 'Neyman'):
            compute_chi_fct = lambda g: \
            np.nansum(self._compute_error_Neyman(g)
            )
        # Pearson
        elif (chi_stat == 'Pearson'):
            compute_chi_fct = lambda g: \
            np.nansum(self._compute_error_Pearson(g)
            )
        else:
            raise Exception("chi_stat is not Valid, should be %s" %(str(self.CHISTAT_VALID)))
        #returns function
        return compute_chi_fct

    def myMCMC(self, max_iterations, random_amplitude, sampling_method='Cauchy', min_acceptance_rate=10 ):
        """

        :param max_iterations:
        :param random_scale:
        :param should_guess_flags:
        :param sampling_method:
        :param  min_acceptance_rate: float [defailt = 5]
            In %, the acceptance rate is the number of useful iterations divided
            by the total number of iterations.
            If it gets below the `failure_rate` treshold, iterations stop.

        :return:
        """


        self.logger.info("Random amplitude : %s", str(random_amplitude))


        # Some useful vars

        count_alpha_ok = 0
        count_min = 0
        total_iterations = 0
        ki_min = sys.maxsize  # that way, it can only go down

        galaxy = self.initial_parameters
        galaxy_old = galaxy.copy()
        galaxy_new = galaxy_old.copy()
        chi_old = self.compute_chi(galaxy)
        #lnprob = self.lnprob(galaxy)

        self.logger.critical("Running GalPak MHSampler with Sampling %s" % (sampling_method))

        names_param = list(galaxy.names)
        names_param.append('reduced_chi')
        # names_param = [
        #   'x', 'y', 'z', 'flux', 'radius', 'inclination', 'pa',
        #   'turnover_radius', 'maximum_velocity', 'velocity_dispersion',
        #   'reduced_chi',
        # ]
        # types_param = ['float32' for i in range(dim_p + 1)]
        #dt = zip(names_param, types_param)
        #chain = np.zeros(max_iterations, dtype=dt)  # intensive operation !
        # chain = Table(names=names_param,dtype=types_param)
        self.chain_rows = []

        rate = 100.
        while (rate > min_acceptance_rate) and (count_alpha_ok < max_iterations):

            # Update loop conditions
            total_iterations += 1

            # Cauchy jumping
            galaxy_new = self._sampling_method(sampling_method, galaxy_old, random_amplitude)


            #new way: from emceee
            #newlnprob = self.lnprob(galaxy_new)
            #diff = 0.5 * (newlnprob - lnprob)

            #if diff < 0:
            #    diff = np.exp(diff) - np.random.rand()

            #if diff > 0:
            #    lnprob = newlnprob
            #    galaxy = galaxy_new

            # Make sure we're inside the preset boundaries
            in_boundaries = (galaxy_new >= self.min_boundaries) * (galaxy_new <= self.max_boundaries)
            if in_boundaries.all():

                # Computing convolved model with new parameters
                chi_new = self.compute_chi(galaxy_new)
                if chi_new<1e-3:
                    #print chi_new,galaxy_new
                    self.logger.error('Chi2 is 0, quit')
                    exit()

                # Compute ratio of likelihood a posteriori
                # exp( old-new / 2 )
                likelihood =  safe_exp(-0.5 * (chi_new - chi_old))

                # likelihood > 1 => new is better
                # likelihood < 1 => accept or reject
                alpha = np.random.rand()

                # Conservation test
                if alpha < likelihood:
                    count_alpha_ok += 1
                    rate = count_alpha_ok * 100. / total_iterations

                    # Save minimum
                    if ki_min > chi_new:
                        ki_min = chi_new
                        count_min = count_alpha_ok

                    galaxy_old = galaxy_new
                    chi_old = chi_new

                    # vect = np.append(galaxy_new, chi_new / self.Ndegree)
                    #chain[count_alpha_ok - 1] = np.float32(vect)
                    # chain.add_row(vect)
                    self.chain_rows.append(list(galaxy_new) +
                                      [chi_new / self.Ndegree])

                    if self.verbose:
                        # Check that parameters are not too close to the boundaries
                        too_close_to_max = np.abs((galaxy_new - self.max_boundaries) / self.max_boundaries) < self.eps
                        too_close_to_min = np.abs((galaxy_new - self.min_boundaries) / self.min_boundaries) < self.eps
                        too_close_to_boundaries = too_close_to_min + too_close_to_max
                        #
                        info = "{count:5d} MIN={count_min:5d} {rate:2.0f}% " \
                               "χ²={ki:3.6f}>{ki_min:3.6f} {params:s}"
                        info = info.format(
                            count=count_alpha_ok,
                            count_min=count_min,
                            ki=chi_new / self.Ndegree,
                            ki_min=ki_min / self.Ndegree,
                            rate=rate,
                            params=galaxy_new.colored_info(too_close_to_boundaries)
                        )
                        if self.model.redshift is not None:
                            mdyn = self.model.compute_Mdyn_at_Rhalf(galaxy_new)
                            info += "log_mdyn={mass:.2f} "
                            info = info.format(mass=np.log10(mdyn))
                        info += "lnlog={lnlog:6.4f} "
                        info = info.format(lnlog=self.loglike(galaxy_new))
                        print(info)

        if len(self.chain_rows)>0:
            chain = Table(names=names_param, rows=self.chain_rows)
        else:
            raise Exception("No useful iterations")

        # Raise if no useful iteration was run
        if rate == 0.:
            self.logger.debug("Last Galaxy Parameters : %s", galaxy_new)
            raise RuntimeError("No useful iteration was run. "
                               "Try with higher max_iterations?")


        # Report
        self.logger.info("Iterations report : %d Total, %d OK, %d%% Rate",
                         total_iterations, count_alpha_ok, rate)
        self.logger.info("Storing results as parameters...")

        # Acceptance Rate
        self.acceptance_rate = rate
        self.logger.info("self.acceptance_rate : useful iterations count / total iterations count : %s " % (self.acceptance_rate) )

        return chain


    #PRIVATE METHODS
    def _sampling_method(self, sampling, params, scale):
        """
        set proposal distribution with sampling
        sampling: 'Cauchy' [default] | 'Normal'
            Cauchy: Lorentz proposal
            Normal: Gaussian proposal
        """
        if sampling == 'Cauchy':
            random_uniforms = np.random.uniform(-np.pi / 2., np.pi / 2., size=len(params))
            galaxy_new = params + scale * np.tan(random_uniforms)
        elif sampling == 'Normal':
            galaxy_new = np.random.normal(params, scale, size=len(params))
        elif sampling == 'AdaptiveCauchy':
            random_uniforms = np.random.uniform(-np.pi / 2., np.pi / 2., size=len(params))
            #@fixme should be covariance
            if len(self.chain_rows)>1500:
                scale = np.sqrt(np.std(np.array(self.chain_rows)[-750:,:-1], axis=0) )**2 / len(params)
            elif len(self.chain_rows) > 500:
                scale = np.sqrt(np.std(np.array(self.chain_rows)[-250:, :-1], axis=0) * 1.25 ) ** 2 / len(params)
            elif len(self.chain_rows) > 50:
                scale = np.sqrt(np.std(np.array(self.chain_rows)[-50:, :-1], axis=0) * 1.5) ** 2 / len(params)

            galaxy_new = params + scale * np.tan(random_uniforms)
        else:
            raise Exception("Not implemented. Should be %s" %(self.SAMPLING_VALID))

        return galaxy_new

    def _init_sampling_scale(self, random_scale, should_guess_flags):
        dim_d = np.size(self.cube.data)
        dim_p = len(self.galaxy)

        # Tweak the random amplitude vector (Kp coeff, as pid)
        # that we can document Model.setup_random_amplitude() adequately
        random_amplitude = np.sqrt(
            (self.min_boundaries - self.max_boundaries) ** 2 / 12.
        ) * dim_p / dim_d

        # Let the model adjust the random amplitude of the parameter jump
        self.model.setup_random_amplitude(random_amplitude)

        # Scale MCMC if needed // allowing vectors
        if random_scale is not None:
            if np.size(random_scale) != 1:
                merge_where_nan(random_scale, np.ones_like(random_amplitude))
            random_amplitude = random_amplitude * random_scale

        # Zero random amplitude where parameters are known
        random_amplitude = random_amplitude * should_guess_flags

        return random_amplitude

    def _compute_error_gauss(self, galaxy):
        """
        It computes the difference between the measured cube
        and the computed cube from given galaxy parameters.
        returns ( D - M ) / stdev
        """

        # Compute convolved cube for given galaxy parameters
        cube_convolved = self.create_convolved_cube(galaxy, self.cube.shape)

        # Diff. between measured cube and convolved cube, scaled by the error
        self.variance_chi = self.variance_cube
        difference = (self.cube - cube_convolved)**2 / self.variance_chi

        return np.ndarray.flatten(difference.data)

    def _compute_error_Mighell(self, galaxy):
        """
            Modified Statistic Mighell 1998
            Sum ( D + min(1,D) - M)^2 / D + 1
        """

        # Compute convolved cube for given galaxy parameters
        cube_convolved = self.create_convolved_cube(galaxy, self.cube.shape)

        tmp_cube = HyperCube(np.where(self.cube.data >= 1, 1., self.cube.data))
        self.variance_chi = self.cube + 1.0
        difference = (self.cube + tmp_cube - cube_convolved ) **2 \
                     / self.variance_chi


        return np.ndarray.flatten(difference.data)

    def _compute_error_Neyman(self, galaxy):
        """
            Modified Neyman statistic
            Humphrey 2009, http://adsabs.harvard.edu/abs/2009ApJ...693..822H
            Sum ( M - D )^2 / max(D,1)
        """

        # Compute convolved cube for given galaxy parameters
        cube_convolved = self.create_convolved_cube(galaxy, self.cube.shape)

        self.variance_chi = HyperCube(np.where(self.cube.data <= 1, 1., self.cube.data))
        difference = (cube_convolved - self.cube)**2 / self.variance_chi

        return np.ndarray.flatten(difference.data)

    def _compute_error_Pearson(self, galaxy):
        """
            Pearson statistic
            Humphrey 2009, http://adsabs.harvard.edu/abs/2009ApJ...693..822H
            Sum ( M - D )^2 / M
         """

        # Compute convolved cube for given galaxy parameters
        cube_convolved = self.create_convolved_cube(galaxy, self.cube.shape)
        difference = (cube_convolved - self.cube)**2 / (cube_convolved)
        self.variance_chi = cube_convolved

        return np.ndarray.flatten(difference.data)

    def _compute_error_cstat(self, galaxy):
        """
         Cash statistique for Poisson noise
         Humphrey 2009, http://adsabs.harvard.edu/abs/2009ApJ...693..822H
         Sum ( M - D + D * log(D/M) )
        """

        # Compute convolved cube for given galaxy parameters
        cube_convolved = self.create_convolved_cube(galaxy, self.cube.shape)

        tmp_cube = np.log(self.cube.data/cube_convolved)
        difference = (cube_convolved - self.cube) + self.cube * tmp_cube
        self.variance_chi = np.ones_like(self.cube.data)

        return np.ndarray.flatten(difference.data)

#########################
######EMCEE 3.0
########################
try:
    from emcee.moves import MHMove

    class CauchyMove(MHMove):
        """A Metropolis step with a Gaussian proposal function.

        Args:
            cov: The covariance of the proposal function. This can be a scalar,
                vector, or matrix and the proposal will be assumed isotropic,
                axis-aligned, or general respectively.
            mode (Optional): Select the method used for updating parameters. This
                can be one of ``"vector"``, ``"random"``, or ``"sequential"``. The
                ``"vector"`` mode updates all dimensions simultaneously,
                ``"random"`` randomly selects a dimension and only updates that
                one, and ``"sequential"`` loops over dimensions and updates each
                one in turn.
            factor (Optional[float]): If provided the proposal will be made with a
                standard deviation uniformly selected from the range
                ``exp(U(-log(factor), log(factor))) * cov``. This is invalid for
                the ``"vector"`` mode.

        Raises:
            ValueError: If the proposal dimensions are invalid or if any of any of
                the other arguments are inconsistent.

        """
        def __init__(self, cov, mode="vector", factor=None):
            # Parse the proposal type.
            try:
                float(cov)

            except TypeError:
                cov = np.atleast_1d(cov)
                if len(cov.shape) == 1:
                    # A diagonal proposal was given.
                    ndim = len(cov)
                    proposal = _diagonal_proposal(np.sqrt(cov), factor, mode)

                elif len(cov.shape) == 2 and cov.shape[0] == cov.shape[1]:
                    # The full, square covariance matrix was given.
                    ndim = cov.shape[0]
                    proposal = _proposal(cov, factor, mode)

                else:
                    raise ValueError("Invalid proposal scale dimensions")

            else:
                # This was a scalar proposal.
                ndim = None
                proposal = _isotropic_proposal(np.sqrt(cov), factor, mode)

            super(CauchyMove, self).__init__(proposal, ndim=ndim)


    class _isotropic_proposal(object):

        allowed_modes = ["vector", "random", "sequential"]

        def __init__(self, scale, factor, mode):
            self.index = 0
            self.scale = scale
            if factor is None:
                self._log_factor = None
            else:
                if factor < 1.0:
                    raise ValueError("'factor' must be >= 1.0")
                self._log_factor = np.log(factor)

            if mode not in self.allowed_modes:
                raise ValueError(("'{0}' is not a recognized mode. "
                                  "Please select from: {1}")
                                 .format(mode, self.allowed_modes))
            self.mode = mode

        def get_factor(self, rng):
            if self._log_factor is None:
                return 1.0
            return np.exp(rng.uniform(-self._log_factor, self._log_factor))

        def get_updated_vector(self, rng, x0):
            return x0 + self.get_factor(rng) * self.scale * rng.randn(*(x0.shape))

        def __call__(self, x0, rng):
            nw, nd = x0.shape
            xnew = self.get_updated_vector(rng, x0)
            if self.mode == "random":
                m = (range(nw), rng.randint(x0.shape[-1], size=nw))
            elif self.mode == "sequential":
                m = (range(nw), self.index % nd + np.zeros(nw, dtype=int))
                self.index = (self.index + 1) % nd
            else:
                return xnew, np.zeros(nw)
            x = np.array(x0)
            x[m] = xnew[m]
            return x, np.zeros(nw)


    class _diagonal_proposal(_isotropic_proposal):

        def get_updated_vector(self, rng, x0):
            random_uniforms = np.random.uniform(-np.pi / 2., np.pi / 2., size=len(self.scale))
            return x0 + self.get_factor(rng) * self.scale * np.tan(random_uniforms)
            #return x0 + self.get_factor(rng) * self.scale * rng.randn(*(x0.shape))

    class _proposal(_isotropic_proposal):

        allowed_modes = ["vector"]

        def get_updated_vector(self, rng, x0):
            random_uniforms = np.random.uniform(-np.pi / 2., np.pi / 2., size=len(self.scale))
            return x0 + self.get_factor(rng) * self.scale * np.tan(random_uniforms)
            #return x0 + self.get_factor(rng) * rng.multivariate_normal(
            #    np.zeros(len(self.scale)), self.scale)

except:
    pass



# coding=utf-8
from __future__ import division

## GENERAL PACKAGES
import os, sys
import logging
import unittest

import numpy
sys.path.append('../')


## GALPAK PACKAGES -- galpak is served both standalone and as part of mpdaf

from galpak import HyperspectralCube

from tests.tools.numpy_ndarray_assertions import NumpyNdArrayAssertions

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('HyperspectralCubeTest')


class HyperspectralCubeTest(unittest.TestCase, NumpyNdArrayAssertions):

    longMessage = True

    root_folder = os.path.abspath(os.path.join(
        os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir
    ))
    fits_folder = os.path.join(root_folder, 'data/input')

    fits_muse_filename = os.path.join(fits_folder, 'cube/subcube_MUSE.fits')
    fits_test_filename = os.path.join(fits_folder, 'cube/subcube_MUSE.fits')

    def test_init_empty_cube(self):
        cube = HyperspectralCube()
        self.assertTrue(cube.is_empty())

    def test_init_with_unfit_filename(self):
        with self.assertRaises(IOError):
            cube = HyperspectralCube.from_file('this_is_not_a_fits_file')

    def test_init_with_all_of_our_input_cube_fits(self):
        for filename in os.listdir(os.path.join(self.fits_folder, 'cube')):
            filepath = os.path.join(self.fits_folder, filename)
            if os.path.isfile(filepath) and filepath.endswith('.fits'):
                # This should not throw (that IS our assertion)
                cube = HyperspectralCube.from_file(filepath, verbose=True)
                log.info(filename)
                log.info("\n"+cube.header.tostring("\n").strip())

    def test_shape_property(self):
        data_a = numpy.array([
            [[0, 1], [2, 3], [-1, +0]],
            [[1, 0], [3, 2], [+3, +2]],
            [[1, 1], [2, 2], [-3, +0]],
        ], dtype=float)
        cube_a = HyperspectralCube(data=data_a)

        self.assertArrayEqual((3, 3, 2), cube_a.shape,
                              "cube.shape should return the shape")

    def test_arithmetic(self):
        inf = numpy.inf
        nan = numpy.nan

        # BRAIN COMPUTED TEST DATA
        data_a = numpy.array([
            [[0, 1], [2, 3], [-1, +0]],
            [[1, 0], [3, 2], [+3, +2]],
            [[1, 1], [2, 2], [-3, +0]],
        ], dtype=float)
        data_b = numpy.array([
            [[1, 1], [0, 0], [+1, -1]],
            [[0, 0], [3, 1], [+3, -2]],
            [[2, 2], [1, 2], [-1, +0]],
        ], dtype=float)
        data_i = numpy.array([
            [[2, 0], [1, 1], [+1, -1]],
        ], dtype=float)
        data_a_plus_1 = numpy.array([
            [[1, 2], [3, 4], [+0, +1]],
            [[2, 1], [4, 3], [+4, +3]],
            [[2, 2], [3, 3], [-2, +1]],
        ], dtype=float)
        data_a_plus_b = numpy.array([
            [[1, 2], [2, 3], [+0, -1]],
            [[1, 0], [6, 3], [+6, -0]],
            [[3, 3], [3, 4], [-4, +0]],
        ], dtype=float)
        data_a_plus_i = numpy.array([
            [[2, 1], [3, 4], [+0, -1]],
            [[3, 0], [4, 3], [+4, +1]],
            [[3, 1], [3, 3], [-2, -1]],
        ], dtype=float)
        data_a_minus_1 = numpy.array([
            [[-1, +0], [1, 2], [-2, -1]],
            [[+0, -1], [2, 1], [+2, +1]],
            [[+0, +0], [1, 1], [-4, -1]],
        ], dtype=float)
        data_1_minus_a = numpy.array([
            [[+1, +0], [-1, -2], [+2, +1]],
            [[+0, +1], [-2, -1], [-2, -1]],
            [[+0, +0], [-1, -1], [+4, +1]],
        ], dtype=float)
        data_a_times_2 = numpy.array([
            [[0, 2], [4, 6], [-2, +0]],
            [[2, 0], [6, 4], [+6, +4]],
            [[2, 2], [4, 4], [-6, +0]],
        ], dtype=float)
        data_a_times_b = numpy.array([
            [[0, 1], [0, 0], [-1, +0]],
            [[0, 0], [9, 2], [+9, -4]],
            [[2, 2], [2, 4], [+3, +0]],
        ], dtype=float)
        data_a_div_2 = numpy.array([
            [[.0, .5], [1.0,  3./2], [-.5,   0]],
            [[.5, .0], [3./2,  1.0], [3./2,  1]],
            [[.5, .5], [1.0,   1.0], [-3./2, 0]],
        ], dtype=float)
        data_a_div_b = numpy.array([
            [[0,     1], [inf, inf], [-1,  -0.]],
            [[inf, nan], [1,     2], [+1,  -1.]],
            [[.5,   .5], [2,     1], [+3,  nan]],
        ], dtype=float)
        data_a_pow_2 = numpy.array([
            [[0, 1], [4, 9], [+1, +0]],
            [[1, 0], [9, 4], [+9, +4]],
            [[1, 1], [4, 4], [+9, +0]],
        ], dtype=float)
        data_2_pow_a = numpy.array([
            [[1, 2], [4, 8], [0.5,  1]],
            [[2, 1], [8, 4], [8,    4]],
            [[2, 2], [4, 4], [1./8, 1]],
        ], dtype=float)
        data_b_pow_a = numpy.array([
            [[1, 1], [0,  0], [+1, +1]],
            [[0, 1], [27, 1], [27, +4]],
            [[2, 2], [1,  4], [-1, +1]],
        ], dtype=float)

        self.assertArrayEqual(data_a_div_b, data_a_div_b,
                              "Sanity check with inf and nan")

        cube_a = HyperspectralCube(data=data_a)
        cube_b = HyperspectralCube(data=data_b)
        cube_i = HyperspectralCube(data=data_i)  # image
        cube_e = HyperspectralCube()             # empty

        # CUBE + NUMBER
        cube_a_plus_1 = cube_a + 1.
        self.assertIsInstance(cube_a_plus_1, HyperspectralCube,
                              "Addition result is a HyperspectralCube")
        self.assertArrayEqual(data_a_plus_1, cube_a_plus_1.data,
                              "Supports addition of cube and number using +")

        # NUMBER + CUBE
        cube_a_plus_1 = 1. + cube_a
        self.assertIsInstance(cube_a_plus_1, HyperspectralCube,
                              "Addition result is a HyperspectralCube")
        self.assertArrayEqual(data_a_plus_1, cube_a_plus_1.data,
                              "Supports addition of number and cube using +")

        # CUBE A + CUBE B
        cube_a_plus_b = cube_a + cube_b
        self.assertIsInstance(cube_a_plus_b, HyperspectralCube,
                              "Addition result is a HyperspectralCube")
        self.assertArrayEqual(data_a_plus_b, cube_a_plus_b.data,
                              "Supports addition of two cubes using +")

        # CUBE A + NDARRAY
        cube_a_plus_b = cube_a + data_b
        self.assertIsInstance(cube_a_plus_b, HyperspectralCube,
                              "Addition result is a HyperspectralCube")
        self.assertArrayEqual(data_a_plus_b, cube_a_plus_b.data,
                              "Supports addition of cube and ndarray using +")

        # CUBE + IMAGE
        cube_a_plus_i = cube_a + cube_i
        self.assertIsInstance(cube_a_plus_i, HyperspectralCube,
                              "Addition result is a HyperspectralCube")
        self.assertArrayEqual(data_a_plus_i, cube_a_plus_i.data,
                              "Supports addition of cube and image using +")

        # IMAGE + CUBE
        cube_a_plus_i = cube_i + cube_a
        self.assertIsInstance(cube_a_plus_i, HyperspectralCube,
                              "Addition result is a HyperspectralCube")
        self.assertArrayEqual(data_a_plus_i, cube_a_plus_i.data,
                              "Supports addition of image and cube using +")

        # ADDITION EXPECTED ERRORS
        with self.assertRaises(TypeError):
            cube_a + cube_e  # right-hand empty cube
        with self.assertRaises(TypeError):
            cube_e + cube_b  # left-hand empty cube
        with self.assertRaises(TypeError):
            cube_b + 'rock'  # left-hand "garbage"
        with self.assertRaises(TypeError):
            '666.' + cube_a  # right-hand "garbage"

        # CUBE - NUMBER
        cube_a_minus_1 = cube_a - 1
        self.assertIsInstance(cube_a_minus_1, HyperspectralCube,
                              "Subtraction result is a HyperspectralCube")
        self.assertArrayEqual(data_a_minus_1, cube_a_minus_1.data,
                              "Supports subtraction of number from cube using -")

        # NUMBER - CUBE
        cube_1_minus_a = 1 - cube_a
        self.assertIsInstance(cube_1_minus_a, HyperspectralCube,
                              "Subtraction result is a HyperspectralCube")
        self.assertArrayEqual(data_1_minus_a, cube_1_minus_a.data,
                              "Supports subtraction of cube from number using -")

        # CUBE * NUMBER
        cube_a_times_2 = cube_a * 2
        self.assertIsInstance(cube_a_times_2, HyperspectralCube,
                              "Multiplication result is a HyperspectralCube")
        self.assertArrayEqual(data_a_times_2, cube_a_times_2.data,
                              "Supports multiplication of cube and number using *")

        # NUMBER * CUBE
        cube_a_times_2 = 2 * cube_a
        self.assertIsInstance(cube_a_times_2, HyperspectralCube,
                              "Multiplication result is a HyperspectralCube")
        self.assertArrayEqual(data_a_times_2, cube_a_times_2.data,
                              "Supports multiplication of number and cube using *")

        # CUBE A * CUBE B
        cube_a_times_b = cube_a * cube_b
        self.assertIsInstance(cube_a_times_b, HyperspectralCube,
                              "Multiplication result is a HyperspectralCube")
        self.assertArrayEqual(data_a_times_b, cube_a_times_b.data,
                              "Supports multiplication of two cubes using *")

        # CUBE / NUMBER
        cube_a_div_2 = cube_a / 2.
        self.assertIsInstance(cube_a_div_2, HyperspectralCube,
                              "Division result is a HyperspectralCube")
        self.assertArrayEqual(data_a_div_2, cube_a_div_2.data,
                              "Supports division of cube and number using /")

        # CUBE / CUBE
        cube_a_div_b = cube_a / cube_b
        self.assertIsInstance(cube_a_div_b, HyperspectralCube,
                              "Division result is a HyperspectralCube")
        self.assertArrayEqual(data_a_div_b, cube_a_div_b.data,
                              "Supports division of two cubes using /")

        # CUBE ** NUMBER
        cube_a_pow_2 = cube_a ** 2
        self.assertIsInstance(cube_a_pow_2, HyperspectralCube,
                              "Exponentiation result is a HyperspectralCube")
        self.assertArrayEqual(data_a_pow_2, cube_a_pow_2.data,
                              "Supports exponentiation of cube by number using **")

        # NUMBER ** CUBE
        cube_2_pow_a = 2 ** cube_a
        self.assertIsInstance(cube_2_pow_a, HyperspectralCube,
                              "Exponentiation result is a HyperspectralCube")
        self.assertArrayEqual(data_2_pow_a, cube_2_pow_a.data,
                              "Supports exponentiation of number by cube using **")

        # CUBE A ** CUBE B
        cube_b_pow_a = cube_b ** cube_a
        self.assertIsInstance(cube_b_pow_a, HyperspectralCube,
                              "Exponentiation result is a HyperspectralCube")
        self.assertArrayEqual(data_b_pow_a, cube_b_pow_a.data,
                              "Supports exponentiation of cube by cube using **")

    def test_slicing_and_indexing(self):
        data_a = numpy.array([
            [[0, 1], [2, 3], [-1, 0]],
            [[1, 0], [3, 2], [+3, 2]],
            [[1, 1], [2, 2], [-3, 0]],
        ], dtype=float)

        cube_a = HyperspectralCube(data=data_a)

        # SLICING USING [:,:,:]
        cube_t = cube_a[0:1, :, :-1]
        data_t = numpy.array([[[0.], [2.], [-1.]]], dtype=float)
        self.assertIsInstance(cube_t, HyperspectralCube,
                              "Truncation result is a HyperspectralCube")
        self.assertArrayEqual(data_t, cube_t.data,
                              "Supports extraction of sub-cube using [:,:,:]")

        # SLICING [min:max] <=> [min:max,:,:]
        cube_t = cube_a[0:1]
        data_t = numpy.array([[[0, 1], [2, 3], [-1, 0]]], dtype=float)
        self.assertArrayEqual(data_t, cube_t.data,
                              "Unspecified axes are defaulted to `:` in slicing")

        # TODO : SLICING UPDATES HEADERS
        # Needs to create test cubes

        # MUTATING ONE VALUE USING [λ,y,x] INDICES
        data_t = numpy.array([
            [[0, 1], [2, 3], [-1, 0]],
            [[1, 0], [3, 9], [+3, 2]],
            [[1, 1], [2, 2], [-3, 0]],
        ], dtype=float)
        cube_t = cube_a.copy()
        cube_t[1, 1, 1] = 9
        self.assertArrayEqual(data_t, cube_t.data,
                              "Supports mutation cube[λ,y,x] = n")

        # MUTATING MULTIPLE VALUES USING [:,y,x] INDICES
        data_t = numpy.array([
            [[0, 1], [2, 9], [-1, 0]],
            [[1, 0], [3, 9], [+3, 2]],
            [[1, 1], [2, 9], [-3, 0]],
        ], dtype=float)
        cube_t = cube_a.copy()
        cube_t[:, 1, 1] = 9
        self.assertArrayEqual(data_t, cube_t.data,
                              "Supports mutation cube[:,y,x] = n")

    #def test_getting_steps(self):
        #cube = HyperspectralCube.from_file(self.fits_test_filename)

        # We merely test the success/failure of the API
        #self.assertArrayEqual(cube.get_steps(), [cube.get_step(0), cube.get_step(1), cube.get_step(2)],
        #                      "cube.get_step(axis) == cube.get_steps()[axis]")

        # Value-checking should be done by the test_compatibility routine
        # self.assertEqual(cube.get_step(0), 0.000194999995074)  # this is a poor test

    def test_wavelength_conversion(self):
        cube = HyperspectralCube.from_file(self.fits_test_filename)
        pixel_index = 3
        self.assertEqual(round(cube.pixel_from_lambda(cube.wavelength_of(pixel_index))), pixel_index,
                         "Wavelength conversion from pixel index to λ and from λ to pixel index")

    def test_write_to_file(self):
        cube = HyperspectralCube.from_file(self.fits_test_filename)
        fits_out_filename = os.path.join(self.fits_folder, 'tmp_output.fits')

        self.assertFalse(os.path.isfile(fits_out_filename),
                         "Sanity check : Output FITS file should not exist ; "
                         "Please remove it by hand and then re-run this test : %s" % fits_out_filename)
        cube.write_to(fits_out_filename)
        self.assertTrue(os.path.isfile(fits_out_filename),
                        "Output FITS file should be created")

        with self.assertRaises(IOError):
            cube.write_to(fits_out_filename)  # clobber option should be false by default

        cube.write_to(fits_out_filename, overwrite=True)  # overwrites without raising
        os.remove(fits_out_filename)  # cleanup



__version__ = '1.30.2'
# Please follow the semver guidelines : http://semver.org
# Many parts of the galpak package use this version number :
# - galpak.__version__ itself, obviously
# - the website generator
# - the documentation generator (sphinx)
# If you need to use this version number with another language than python,
# you'll have to write a parser. galpak needs to integrate into MPDAF
# whose setup prevents the usage of the preferred VERSION file.



# This file is the tasks repository for `doit`.
# See http://pydoit.org/tasks.html#intro

# Simply run `doit` from the projects directory to build the website.

import sys
import os, io
from os import listdir
from os.path import isfile, join, basename, abspath
import tarfile
import shutil
import re

from flask_frozen import Freezer

sys.path.insert(1, 'web')

from web.run import app, downloads_path

# sorting arrays
def sorted_aphanumeric(data):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
    return sorted(data, key=alphanum_key)

__version__ = None
f = open('VERSION')
exec(f.read())
print("Galpak version",__version__)

   

#### TASKS #####################################################################

DOIT_CONFIG = {'default_tasks': ['website']}


def task_doc():
    """
    Generate the documentation webpages using sphinx.
    """
    return {
        'actions': [build_doc],
        'verbosity': 2
    }


def task_tarball():
    """
    Generate a tarball of module, ready for download.
    """
    return {
        'actions': [build_tarball],
        'verbosity': 2
    }


def task_website():
    """
    Re-build the whole website. (documentation pages included)
    """
    return {
        'actions': [build_doc, build_website],
        'verbosity': 2
    }


def task_publish():
    """
    Upload the website to the remote webserver.
    """
    return {
        'actions': [publish],
        'verbosity': 2
    }


def task_all():
    """
    Do all the tasks.
    """
    return {
        'actions': [build_doc, build_tarball, build_website, publish],
        'verbosity': 2
    }


def task_test():
    """
    Run the test suite.
    """
    return {
        'actions': ['python -m pytest --ignore=tests/other'],
        'verbosity': 2
    }


### HELPERS ####################################################################

def filter_tarball(tarinfo):
    """
    A filter for the files we do not want in the tarball.
    Return None to exclude, or return the `tarinfo` if okay.
    """
    name = tarinfo.name
    # Remove the module directory name
    i = name.find('/')
    if i > 0:
        name = name[i+1:]
    # Exclude all these
    if name.endswith(".pyc") or \
            name.startswith('web') or \
            name.startswith('data/debug') or \
            name.startswith('data/tests') or \
            name.startswith('build') or \
            name.startswith('.idea') or \
            name.startswith('.git') or \
            name.startswith('slides') or\
            name.startswith('.tox') or\
            name.startswith('.doit') or \
            name.startswith('.pytest') or \
            name.startswith('dist') or \
            name.startswith('MGE') or \
            name.startswith('html') or \
            name.endswith('pymulti') or \
            name.endswith('eggs/') or \
            name.endswith('info'):
        return None
    return tarinfo


def make_tarball(output_filename, sources_dirnames):
    """
    Helper to make `output_filename` a tarball (.tar.gz) file from the files
    list `sources_dirnames`.
    """
    print("Generating tarball %s..." % output_filename)
    with tarfile.open(output_filename, "w:gz") as tar:
        for source_dirname in sources_dirnames:
            tar.add(source_dirname,
                    arcname=basename(source_dirname),
                    filter=filter_tarball)


#### URL GENERATORS ############################################################

# Freezer needs help to generate the downloads filenames
def downloads_filename():
    downloads_files = [f for f in listdir(downloads_path)
                       if isfile(join(downloads_path, f))]
    downloads_files = sorted_aphanumeric(downloads_files)
    downloads_files.reverse()
    downloads_files = downloads_files[0:5]
    print( os.path.abspath(downloads_path) )
    print(downloads_files)
    for tarball in downloads_files:
        yield {'filename': tarball}


#### ACTIONS ###################################################################

def build_doc():
    print("Generating sphinx documentation webpages...")
    os.chdir('doc')
    os.system('make html_silently')
    os.chdir('..')


def build_website():
    """
    Generate static webpages from the flask app.
    """
    print("Generating static webpages...")
    #print("config",app.config)
    app.config['JSON_SORT_KEYS']=False
    freezer = Freezer(app)
    freezer.register_generator(downloads_filename)
    freezer.freeze()

    # Copy sphinx documentation static pages into website's doc/
    source = abspath("doc/build/html")
    target = abspath("web/build/doc")
    print("Copy documentation from %s to %s..." % (source, target))
    shutil.rmtree(target, ignore_errors=True)
    shutil.copytree(source, target)


def build_tarball():
    #from galpak.__version__ import __version__
    # Read version.py

    source = abspath(".")
    target = abspath("build/galpak_"+__version__+".tar.gz")
    make_tarball(target, [source])


def publish():
    print("Upload files to webserver...")
    cmd = "rsync -r --delete --protocol=29 web/build/ nbouche@mistral:galpak/"
    os.system(cmd)

    print("Building wheel")
    cmd = 'rm dist/* -f '
    os.system(cmd)
    cmd = "python3 setup.py sdist bdist_wheel"
    os.system(cmd)
    #https://packaging.python.org/guides/using-testpypi/
    print("Pushing to pypi")
    cmd = "python3 -m twine upload dist/*"
    os.system(cmd)



#!/usr/bin/python3.5

import galpak, sys
import pprofile

gk=galpak.GalPaK3D('data/input/GalPaK_cube_1101_from_paper.fits')
prof = pprofile.Profile()

with prof:
    gk.run_mcmc(max_iterations=5000,random_scale=2, verbose=False)

prof.annotate(sys.stdout)



import numpy as np
import benchmark

#Benchmark Report
#================
#
#Benchmark numpy nan to num
#--------------------------
#
#       name | rank |  runs |      mean |        sd | timesBaseline
#------------|------|-------|-----------|-----------|--------------
#key mapping |    1 | 1e+04 | 0.0001705 |  1.39e-05 |           1.0
# nan to num |    2 | 1e+04 |  0.000952 | 0.0002365 | 5.58281984443
#
#Each of the above 20000 runs were run in random, non-consecutive order by
#`benchmark` v0.1.5 (http://jspi.es/benchmark) with Python 2.7.5+
#Linux-3.11.0-12-generic-x86_64 on 2014-01-27 10:33:55.


class Benchmark_numpy_nan_to_num(benchmark.Benchmark):

    each = 10000  # configure number of runs

    def setUp(self):
        # Can also specify tearDown, eachSetUp, and eachTearDown
        self.input = np.random.rand(32, 32, 32)

    def eachSetUp(self):
        self.a = self.input.copy()

    def test_nan_to_num(self):
        self.a = np.nan_to_num(self.a)

    def test_key_mapping(self):
        self.a[np.isnan(self.a)] = 0


if __name__ == '__main__':
    benchmark.main(format="markdown", numberFormat="%.4g")




# -*- coding: utf-8 -*-

import numpy as np


try:
    from astropy.io import fits
except ImportError:
    import pyfits as fits

import logging

logging.basicConfig(level=logging.INFO)

#
# This file contains the PointSpreadFunction and LineSpreadFunction interfaces
# as well as some basic implementations of these interfaces :
#   - Gaussian PSF
#   - Moffat PSF
#   - Gaussian LSF
#   - MUSE LSF (only if mpdaf module is available)
#
# The instrument will use both 2D PSF and 1D LSF
# to create a 3D PSF with which it will convolve the cubes.
#


## POINT SPREAD FUNCTIONS ######################################################


class PointSpreadFunction:
    """
    This is the interface all Point Spread Functions (PSF) should implement.
    """

    logger = logging.getLogger('GalPaK')

    def as_image(self, for_cube):
        """
        Should return this PSF as a 2D image shaped [for_cube].

        for_cube: HyperspectralCube
            Has additional properties computed and attributed by GalPaK :
                - xy_step (in ")
                - z_step (in µm)
                - z_central (in µm)

        :rtype: ndarray
        """
        raise NotImplementedError()


class NoPointSpreadFunction(PointSpreadFunction):
    """
    A point spread function that does not spread anything, and returns the cube unchanged.
    Passing this to the instrument's psf is the same as passing None.
    """
    def __init__(self):
        pass

    def as_image(self, for_cube):
        """
        Return the identity PSF, chock-full of ones.
        """
        shape = for_cube.shape[1:]
        return np.ones(shape)


class ImagePointSpreadFunction(PointSpreadFunction):
    """
    A custom point spread function using a provided 2D image
    that should have the same shape as the cube's (x,y) and
    centroid should be at
        xo = (shape[1] - 1) / 2 - (shape[1] % 2 - 1)
        yo = (shape[0] - 1) / 2 - (shape[0] % 2 - 1)

    """
    def __init__(self, image_2d):
        """
        accepts fits file or ndarray

        """
        if isinstance(image_2d, str):
            self.filename = image_2d
            my_image = fits.open(image_2d)
            if my_image['PRIMARY'].data is not None:
                 image_2d = my_image['PRIMARY'].data
            elif my_image['DATA'].data is not None:
                 image_2d = my_image['DATA'].data
            
        elif isinstance(image_2d,np.ndarray):
             image_2d = image_2d
             self.filename = str(image_2d.__class__)
        else:
            raise ValueError(' PSF provided is not a fits file nor an array !!')

        logging.info( "Normalizing PSF image")
        self.image_2d =  image_2d /  image_2d.sum()

        if isinstance(self.image_2d,np.ndarray) is False:
            raise ValueError(' PSF provided could not be stored in an ndarray')

        if len(self.image_2d.shape)!=2:
            raise ValueError(' PSF provided is not a 2D image')

    def as_image(self, for_cube):
        #check for size
        if for_cube.shape[1:] != self.image_2d.shape:
            raise ValueError(' PSF Image and cube have different sizes: %s vs. %s' % (str(for_cube.shape[1:]),str(self.image_2d.shape)) )
        return self.image_2d

    def __str__(self):
        return """[PSF] :
  type = custom
  filename = {i.filename}""".format(i=self)


class GaussianPointSpreadFunction(PointSpreadFunction):
    """
    The default Gaussian Point Spread Function.

    fwhm: float
        Full Width Half Maximum in arcsec, aka. "seeing".
    pa: float [default is 0.]
        Position Angle of major-axis, anti-clockwise rotation from Y-axis, in angular degrees.
    ba: float [default is 1.0]
        Axis ratio of the ellipsis, b/a ratio (y/x).
    """
    def __init__(self, fwhm=None, pa=0, ba=1.0):
        self.fwhm = fwhm
        self.pa = pa
        self.ba = ba

    def __str__(self):
        return """[PSF] :
  type = gaussian
  fwhm         = {i.fwhm} "
  pa           = {i.pa} °
  ba           = {i.ba}""".format(i=self)

    def as_image(self, for_cube, xo=None, yo=None):

        shape = for_cube.shape[1:]

        if xo is None:
            xo = (shape[1] - 1) / 2 - (shape[1] % 2 - 1)
        if yo is None:
            yo = (shape[0] - 1) / 2 - (shape[0] % 2 - 1)

        y, x = np.indices(shape)
        r = self._radius(xo, yo, x, y)
        fwhm = self.fwhm / for_cube.xy_step #in pixels

        psf = np.exp(-0.5 * (r / (fwhm / 2.35482)) ** 2)

        return psf / psf.sum()

    def _radius(self, xo, yo, x, y):
        """
        Computes the radii, taking into account the variance and the elliptic shape
        """
        dx = xo - x
        dy = yo - y
        # Rotation matrix around z axis
        # R(90)=[[0,1],[-1,0]] so anti-clock-wise y -> x & x -> -y
        radian_pa = np.radians(self.pa)
        dx_p = dx * np.cos(radian_pa) + dy * np.sin(radian_pa)
        dy_p = -dx * np.sin(radian_pa) + dy * np.cos(radian_pa)

        return np.sqrt(dx_p ** 2 / self.ba ** 2+ dy_p ** 2 )


class MoffatPointSpreadFunction(GaussianPointSpreadFunction):
    """
    The Moffat Point Spread Function

    fwhm if alpha is None: float [in arcsec]
        Moffat's distribution fwhm variable : http://en.wikipedia.org/wiki/Moffat_distribution
    alpha if fwhm is None: float [in arcsec]
        Moffat's distribution alpha variable : http://en.wikipedia.org/wiki/Moffat_distribution
    beta: float
        Moffat's distribution beta variable : http://en.wikipedia.org/wiki/Moffat_distribution

    pa: float [default is 0.]
        Position Angle of major-axis, the anti-clockwise rotation from Y,
        in angular degrees.
    ba: float [default is 1.0]
        Axis ratio of the ellipsis, b/a ratio (y/x).
    """

    def __init__(self, fwhm=None, alpha=None, beta=None, pa=None, ba=None):
        self.alpha = alpha
        self.beta = beta
        if (alpha is None) or (fwhm is None):
            GaussianPointSpreadFunction.__init__(self, fwhm, pa, ba)
        else:
            raise Exception("Moffat error: please set alpha or fwhm but not both ")
        if self.pa is None:
            raise Exception("Moffat error: please set P.A. 'pa'")
        if self.ba is None:
            raise Exception("Moffat error: please set axis ratio b/a 'ba'")
        if self.beta is None:
            raise Exception("Moffat error: please set beta parameter")

    def __str__(self):
        return """[PSF] :
  type = Moffat
  fwhm         = {i.fwhm} "
  alpha        = {i.alpha} "
  beta         = {i.beta}  
  pa           = {i.pa} °
  ba           = {i.ba}""".format(i=self)

    def as_image(self, for_cube, xo=None, yo=None):
       
        shape = for_cube.shape[1:]

        if xo is None:
            xo = (shape[1] - 1) / 2 - (shape[1] % 2 - 1)
        if yo is None:
            yo = (shape[0] - 1) / 2 - (shape[0] % 2 - 1)

        y, x = np.indices(shape)
        r = self._radius(xo, yo, x, y)

        beta = self.beta
        if self.alpha is None:
            # Get the FWHM in pixels (we assume the pixels are squares!)
            fwhm = self.fwhm / for_cube.xy_step
            alpha = fwhm / (2.*np.sqrt(2.**(1./beta)-1) )
        if self.fwhm is None:
            # Get the FWHM in pixels (we assume the pixels are squares!)
            alpha = self.alpha / for_cube.xy_step
            fwhm = alpha  * ( 2.*np.sqrt(2.**(1./beta)-1) )  

        psf = (1. + (r / alpha) ** 2) ** (-beta)

        return psf / psf.sum()


## LINE SPREAD FUNCTIONS #######################################################


class LineSpreadFunction:
    """
    This is the interface all Line Spread Functions (LSF) should implement.
    """

    z_cunit = 'Undef'

    def as_vector(self, for_cube):
        """
        Should return this LSF as a 1D vector shaped [for_cube].

        for_cube: HyperspectralCube

        :rtype: ndarray
        """
        raise NotImplementedError()

class NoLineSpreadFunction(LineSpreadFunction):
    """
    A point spread function that does not spread anything, and returns the cube unchanged.
    Passing this to the instrument's lsf is the same as passing None.
    """
    def __init__(self):
        pass

    def as_vector(self, for_cube):
        """
        returns the identity LSF with ones
        :param for_cube:
        :return:
        """
        shape = for_cube.shape[0]
        return np.ones(shape)

    def __str__(self):
        return """[LSF] :\n  type = undefined """


class VectorLineSpreadFunction(LineSpreadFunction):
    """
    A custom line spread function using a provided 1D `vector`
    that should have the same length as the cube's (z).
    Should be centered around zo = (zsize - 1) / 2 - (zsize % 2 - 1)
    """

    def __init__(self, vector):
        self.vector = vector

    def as_vector(self, for_cube):
        return self.vector

    def __str__(self):
        return """[LSF] \n type = Custom """


class GaussianLineSpreadFunction(LineSpreadFunction):
    """
    A line spread function that spreads as a gaussian.
    We assume the centroid is in the middle.

    fwhm: float
        Full Width Half Maximum, in units of CUNIT3
    """
    def __init__(self, fwhm):
        self.fwhm = fwhm

    def __str__(self):
        return """[LSF] :
  type = Gaussian
  fwhm = {i.fwhm}  {i.z_cunit} \n""".format(i=self)

    def as_vector(self, for_cube):
        # Std deviation from FWHM
        sigma = self.fwhm / 2.35482 / for_cube.z_step
        # Resulting vector shape
        depth = for_cube.shape[0]
        # Assymmetric range around 0
        zo = (depth - 1) / 2 - (depth % 2 - 1)
        z_range = np.arange(depth) - zo
        # Compute gaussian (we assume peak is at 0, ie. µ=0)
        lsf_1d = self.gaussian(z_range, 0, sigma)
        # Normalize and serve
        return lsf_1d / lsf_1d.sum()

    @staticmethod
    def gaussian(x, mu, sigma):
        """
        Non-normalized gaussian function.

        x : float|numpy.ndarray
            Input value(s)
        mu : float
            Position of the peak on the x-axis
        sigma : float
            Standard deviation

        :rtype: Float value(s) after transformation, of the same shape as input x.
        """
        return np.exp((x - mu) ** 2 / (-2. * sigma ** 2))


class MUSELineSpreadFunction(LineSpreadFunction):
    """
    A line spread function that uses MPDAF's LSF.
    See http://urania1.univ-lyon1.fr/mpdaf/chrome/site/DocCoreLib/user_manual_PSF.html

    .. warning::
        This requires the mpdaf module.
        Currently, the MPDAF module only works for odd arrays.
    model: string
        See ``mpdaf.MUSE.LSF``'s ``typ`` parameter.
    """
    def __init__(self, model="qsim_v1"):
        self.model = model
        try:
            from mpdaf.MUSE import LSF
        except ImportError:
            raise ImportError("You need the mpdaf module to use MUSELineSpreadFunction.")
        self.lsf = LSF(typ=self.model)


    def __str__(self):
        return """MUSE LSF : model = '{i.model}'""".format(i=self)

    def as_vector(self, cube):
        # Resulting vector shape
        depth = cube.shape[0]
        odd_depth = depth if depth % 2 == 1 else depth+1
        # Get LSF 1D from MPDAF
        if 'micron' in cube.z_cunit.lower():
            wavelength_aa = cube.z_central * 1e4  # unit conversion from microns to AA
            z_step_aa = cube.z_step * 1e4
        else:
            wavelength_aa = cube.z_central
            z_step_aa = cube.z_step

        lsf_1d = self.lsf.get_LSF(lbda=wavelength_aa, step=z_step_aa, size=odd_depth)

        x=np.arange(np.size(lsf_1d))
        sigma = np.sqrt(np.sum(x**2*lsf_1d)-np.sum(x*lsf_1d)**2) * cube.z_step

        self.fwhm = sigma * 2.35

        # That LSF is of an odd depth, truncate it if necessary
        # FIXME @nicolas : this is a hotfix, not really pretty, how can we do this better?
        if depth % 2 == 0:
            lsf_1d = lsf_1d[:-1]
        # Normalize and serve
        return lsf_1d / lsf_1d.sum()



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK: Instrument: ')

from .spread_functions import *
from . import convolution

class Instrument(object):
    """
    This is a generic instrument class to use or extend.

    psf: None|PointSpreadFunction
        The 2D Point Spread Function instance to use in the convolution, or None (the default).
        When this parameter is None, the instrument's PSF will not be applied to the cube.
    lsf: None|LineSpreadFunction
        The 1D Line Spread Function instance to use in the convolution, or None (the default).
        Will be used in the convolution to spread the PSF 2D image through a third axis, z.
        When this parameter is None, the instrument's PSF will not be applied to the cube.
    """

    # SPREAD FUNCTIONS
    psf = None  # Object implementing PointSpreadFunction, or None.
    lsf = None  # Object implementing LineSpreadFunction, or None.
    # MEMORIZATION HOLDERS
    psf3d = None      # 3D ndarray, product of 2D PSF and 1D LSF
    psf3d_fft = None  # Fast Fourier Transform of above

    cube_default_z_step = None   #
    cube_default_z_central = None   #
    cube_default_xy_step = None     #
    cube_default_cunit = 'Undef'  #

    cube_xy_step = cube_default_xy_step
    xy_step = cube_default_xy_step
    z_step = cube_default_z_step
    z_central = cube_default_z_central
    z_cunit = cube_default_cunit
    z_step_kms = None

    logger = logger

    def __init__(self, psf=None, lsf=None):

        if (self.xy_step is None) and (self.cube_default_xy_step is not None):
            self.xy_step = self.cube_default_xy_step

        if (self.z_step is not None) and (self.z_central is not None):
                self.z_step_kms = 3e5 * self.z_step / self.z_central
        else:
                self.z_step_kms = None

        if psf:
            if not isinstance(psf, PointSpreadFunction):
                raise ValueError("PSF should be an instance of galpak.PointSpreadFunction")
            self.psf = psf

        if lsf:
            if not isinstance(lsf, LineSpreadFunction):
                raise ValueError("LSF should be an instance of galpak.LineSpreadFunction")
            self.lsf = lsf
            self.lsf.z_cunit = self.z_cunit

    def use_pixelsize_from_cube(self, cube):
        """
        This should use_pixelsize_from_cube the Instrument and its PSF with the values from the Cube.
        Sometimes an instrument's precision depends upon which region of the sky and spectrum it's looking at.
        This callback is here to solve that problem.

        The passed cube has additional attributes provided by GalPak :
            - xy_step (in ")
            - z_step (in cunit3)
            - z_central (in cunit3)
        """

        #xy_step = cube.get_steps()[1]
        #z_step = cube.header.get('CDELT3')
        #z_cunit = cube.header.get('CUNIT3')
        #crpix = cube.header.get('CRPIX3')
        #crval = cube.header.get('CRVAL3')

        if cube.z_step is None or cube.z_central is None or cube.xy_step is None:
            self.logger.warning("The Cube probably lacks spectral headers.")

        if self.cube_default_xy_step is None:
            # if no default, use cube
            if cube.xy_step is not None:
                self.logger.warning('Forcing instrument pixelscale to that of cube')
                self.cube_xy_step = cube.xy_step
                self.xy_step = cube.xy_step  # pixel size in arcsec
            #if no default and no cube header Error
            else:
                raise ValueError("No default pixscale and No header information")

        elif self.cube_default_xy_step is not None:
            # use default from instrument if cube does not
            if cube.xy_step is not None:
                self.cube_xy_step = cube.xy_step
                if np.abs(self.cube_default_xy_step - self.cube_xy_step) / self.cube_xy_step > 0.1:
                    raise ValueError("INSTRUMENT pixscale differs from that of data!!")
                else:
                    self.xy_step = self.cube_xy_step

        if cube.z_step is not None:
            self.z_step = cube.z_step
        if cube.z_step is not None and cube.z_central is not None:
            self.z_step_kms = 3e5 * cube.z_step / cube.z_central  # km/s
            #logger.info("z_step_kms = %.3e km/s", self.z_step_kms)
        if cube.z_central is not None:
            self.z_central = cube.z_central
        if cube.z_cunit is not None:
            self.z_cunit = cube.z_cunit
            self.lsf.z_cunit = cube.z_cunit  #fixme: need to convert default lsf_fwhm to proper units

    def convolve(self, cube):
        """
        Convolve the provided data cube using the 3D Spread Function
        made from the 2D PSF and 1D LSF.
        Should transform the input cube and return it.
        If the PSF or LSF is None, it will do nothing.

        .. warning::
            The 3D spread function and its fast fourier transform are
            memoized for optimization, so if you change the instrument's
            parameters after a first run, the SF might not reflect your changes.
            Delete ``psf3d`` and ``psf3d_fft`` to clear the memory.

        cube: HyperspectralCube
        """
        # Skip if missing PSF or LSF
        if self.psf is None or self.lsf is None:
            return cube

        # PSF's FFT is memoized for optimization
        if self.psf3d_fft is None:
            # Apply LSF to PSF
            self.psf3d = self.extrude_psf(self.psf.as_image(cube), self.lsf.as_vector(cube))
            cube_convolved, self.psf3d_fft = convolution.convolve_3d_same(cube.data, self.psf3d, compute_fourier=True)
        else:
            cube_convolved, __ = convolution.convolve_3d_same(cube.data, self.psf3d_fft, compute_fourier=False)

        cube.data = cube_convolved

        return cube

    @staticmethod
    def extrude_psf(psf2d, lsf1d):
        """
        Apply the 1D LSF to provided 2D PSF image, and return the resulting 3D PSF cube.
        LSF should extrude the image along the z-axis (wavelength).
        """
        return psf2d * lsf1d[:, None, None]

    def __str__(self):
        return """[INSTRUMENT] :
  type = {i.__class__.__name__}
  pixscale = {i.xy_step} "

{i.psf}
{i.lsf}
cube_xy_step   = {i.cube_xy_step} "
cube_z_step    = {i.z_step} {i.z_cunit}

cube z_step_kms = {i.z_step_kms} km/s  at {i.z_central} {i.z_cunit}

""".format(i=self)


class Generic(Instrument):
    """
    Generic instrument with no defaults.
    """

    def __init__(self, psf=None, lsf=None,
                 lsf_fwhm=None,
                 psf_fwhm=None, psf_pa=None, psf_ba=None,
                 default_spaxel_size=None, default_zstep=None, default_zcentral=None, default_cunit=None):
        """
        Generic instrument with no defaults.

          lsf_fwhm None
            the lsf FWHM in units of the cube CUNIT3
          psf_fwhm None
            the psf FWHM in arcsec
          psf_pa None
            the psf PA of the major-axis, anti-clockwise
          psf_ba None
            the psf axis ratio
         default_spaxel_size
            the Generic Instrument spaxel size [kpc, "]
         default_zstep_size
            the Generic Instrument cdelt3 step in [default_cunit]
         default_cunit
            the Generic Instrument cdelt3 unit
         default_zcentral
            the Generic Instrument central wavelength/frequency in [default_cunits]
        """

        _errmsg = "Generic instrument's `%s` is required."

        if default_spaxel_size is not None:
            self.cube_default_xy_step = default_spaxel_size

        if (default_zcentral is not None and default_cunit is not None and default_zstep is not None):
            if default_zstep is not None:
                self.cube_default_z_step = default_zstep

            if default_zcentral is not None:
                self.cube_default_z_central = default_zcentral

            if default_cunit is not None:
                self.cube_default_cunit = default_cunit
        elif not(default_zcentral is None and default_cunit is None and default_zstep is None):
            self.logger.error("Please specify together the default_zcentral, default_cunit, default_ztep values")


        if lsf is None:
            if lsf_fwhm is None:
                raise TypeError(_errmsg % 'lsf_fwhm')
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            if psf_fwhm is None:
                raise TypeError(_errmsg % 'psf_fwhm')
            if psf_pa is None:
                raise TypeError(_errmsg % 'psf_pa')
            if psf_ba is None:
                raise TypeError(_errmsg % 'psf_ba')
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm,
                                              pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        """
        method to use pixelsize from cube, called by galpak.__init__
        """

        Instrument.use_pixelsize_from_cube(self, cube)


class ALMA(Instrument):
    """
    ALMA instrument mode
    psf_fwhm: float
        The PSF Full Width Half Maximum in arcsec, aka. "seeing".
    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    spax_scale: float
        The spatial scale (arcsec)
    zstep_scale: float
        The frequency pixel size (Hz)

    """
    cube_default_z_step = None   # Hz
    cube_default_z_central = None   # Hz
    cube_default_xy_step = None     # "
    cube_default_cunit = 'Hz'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=None, psf_fwhm=1.2, psf_pa=0., psf_ba=1.0, pixscale=cube_default_xy_step, zstep_scale=cube_default_z_step):
        #if (spaxel_scale is None):
        #        self.logging.WARNING('WARNING: ALMA instrument has no default spatial scale (arcsec), will try to use the cube header. It can be specified with spaxel_scale(arcsec)')
        #if (zstep_scale is None):
        #        self.logging.WARNING('WARNING: ALMA instrument has no default freq. scale (Hz), will try to use the cube header. It can be specified with zstep_scale')

        if pixscale is None:
            self.logger.warning("ALMA: no spaxel defined; will use cube header")

        self.xy_step = self.cube_default_xy_step = pixscale  # "
        self.z_step = zstep_scale
        self.lsf_fwhm = lsf_fwhm

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)

        if (cube.z_step is not None) and (self.lsf.fwhm is None):
            self.lsf.fwhm = cube.z_step
            self.logger.warning('Setting LSF FWHM to 1 z_pixel')

class HARMONI(Instrument):
    """
    pixelscale : float
         4, 10, 20, 30 mas

    """
    VALID_pixscale = [4, 10, 20, 30]

    cube_default_xy_step = None
    cube_default_z_step = None   # µm
    cube_default_z_central = None   # µm
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, pixscale=cube_default_xy_step, psf_fwhm=None, psf_pa=0., psf_ba=1.0, lsf_fwhm=None):
        if pixscale is not None:
            self.logger.warning(" Will assume pixel scale in arcsecs")
            self.xy_step = self.cube_default_xy_step =  pixscale   # @fixme deal with units
        else:
            self.cube_default_xy_step = None
            self.logger.warning("HARMONI: no spaxel defined; will use cube header")

        if lsf is None:
            if lsf_fwhm is None:
                raise TypeError('Harmoni lsf_fwhm is required when no LSF')
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            if psf_fwhm is None:
                raise TypeError('Harmoni psf_fwhm is required when no PSF')
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)
        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)



class MUSE(Instrument):
    """
    MUSE Wide Field Mode (default mode).
    Also used as a parent class for both field modes.

    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    psf_fwhm: float
        The PSF's Full Width Half Maximum in arcsec, aka. "seeing".
        Default to 0.8.

    lsf_mode 'gaussian' for a LSF 1D gaussian // 'mpdaf' qsim version of LSF
    """
    cube_default_xy_step = 0.2      # Spatial step in "
    cube_default_z_step = 1.25   # Spectral step in µm
    cube_default_z_central = 6564  # µm
    cube_default_cunit = 'Angstrom'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=None, psf_fwhm=1.0, psf_pa=0., psf_ba=1.0):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def lsf_fwhm_for_MUSE(self, wave=None):
        """
        computes fwhm of lsf from wavelength using
         fwhm = 5.835 10−8λ2 − 9.080 10−4λ+ 5.983 [Bacon+17]
        :param wave: wavelength in Angstrom
        :return:
        """
        fwhm = None
        if wave is None:
            try:
                wave = self.z_central
                fwhm = 5.835e-8*wave**2-9.08e-4*wave+5.983
                self.logger.info("lsf_fwhm set up for lambda in MUSE. See Bacon+17")
            except:
                self.logger.warning("Wavelength undefined for MUSE LSF\n Specify wave or z_central ")
        else:
            fwhm = 5.835e-8*wave**2-9.08e-4*wave+5.983

        return fwhm

    def use_pixelsize_from_cube(self, cube):
        """
        MUSE's LSF's FWHM depends on the z step
        """

        Instrument.use_pixelsize_from_cube(self, cube)
        try:
            if self.lsf.fwhm is None:
                self.logger.info("Adjusting LSF FWHM for MUSE cube.z_central")
                self.lsf.fwhm = self.lsf_fwhm_for_MUSE()
        except:
            pass

class MUSEWFM(MUSE):
    """
    MUSE Wide Field Mode (default mode).
    See MUSE for its default values.
    """


class MUSENFM(MUSE):
    """
    MUSE Narrow Field Mode.
    Has default values of MUSE, except for the spatial step.
    """
    cube_default_xy_step = 0.025  # Spatial step in "


class KMOS(Instrument):
    cube_default_z_step = 0.0002   # µm
    cube_default_z_central = None   # µm
    cube_default_xy_step = 0.2    # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00065, psf_fwhm=0.6, psf_pa=90., psf_ba=0.8):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)


class SINFOK250(Instrument):
    """
    psf_fwhm: float
        The PSF Full Width Half Maximum in arcsec, aka. "seeing".
    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    cube_default_z_step = 2.45e-4   # µm
    cube_default_z_central = 2.20   # µm
    cube_default_xy_step = 0.125    # "
    cube_default_cunit = 'micron'
    """
    cube_default_z_step = 2.45e-4   # µm
    cube_default_z_central = 2.20   # µm
    cube_default_xy_step = 0.125    # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00065, psf_fwhm=1.2, psf_pa=90., psf_ba=0.8):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)


class SINFOK100(Instrument):
    """
    psf_fwhm: float
        The PSF Full Width Half Maximum in arcsec, aka. "seeing".
    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    cube_default_z_step = 2.45e-4   # µm
    cube_default_z_central = 2.20   # µm
    cube_default_xy_step = 0.05     # "
    cube_default_cunit = 'micron'
    """

    cube_default_z_step = 2.45e-4   # µm
    cube_default_z_central = 2.20   # µm
    cube_default_xy_step = 0.05     # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00065, psf_fwhm=1.2, psf_pa=90., psf_ba=0.8):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)


class SINFOH250(Instrument):
    """
    psf_fwhm: float
        The PSF Full Width Half Maximum in arcsec, aka. "seeing".
    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    cube_default_z_step = 1.95e-4   # µm
    cube_default_z_central = 1.50   # µm
    cube_default_xy_step = 0.125    # "
    cube_default_cunit = 'micron'
    """

    cube_default_z_step = 1.95e-4   # µm
    cube_default_z_central = 1.50   # µm
    cube_default_xy_step = 0.125    # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00078, psf_fwhm=1.2, psf_pa=90., psf_ba=0.8):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)


class SINFOJ250(Instrument):
    """
    psf_fwhm: float
        The PSF Full Width Half Maximum in arcsec, aka. "seeing".
    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    cube_default_z_step = 1.45e-4   # µm
    cube_default_z_central = 1.20   # µm
    cube_default_xy_step = 0.125    # "
    cube_default_cunit = 'micron'
    """

    cube_default_z_step = 1.45e-4   # µm
    cube_default_z_central = 1.20   # µm
    cube_default_xy_step = 0.125    # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00065, psf_fwhm=1.2, psf_pa=90., psf_ba=0.8):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)


class SINFOJ100(Instrument):
    """
    psf_fwhm: float
        The PSF Full Width Half Maximum in arcsec, aka. "seeing".
    psf_pa: float
        The PSF Position Angle (from y-axis) of the major axis: anti-clockwise rotation from Y axis, in angular degrees.
    psf_ba: float
        The PSF axis ratio (<1)

    cube_default_z_step = 1.45e-4   # µm
    cube_default_z_central = 1.20   # µm
    cube_default_xy_step = 0.05     # "
    cube_default_cunit = 'micron'
    """

    cube_default_z_step = 1.45e-4   # µm
    cube_default_z_central = 1.20   # µm
    cube_default_xy_step = 0.05     # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00065, psf_fwhm=1.2, psf_pa=90., psf_ba=0.8):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        Instrument.use_pixelsize_from_cube(self, cube)


class OSIRIS(Instrument):
    cube_default_z_step = 0.25e-4   # µm
    cube_default_z_central = 1.99   # µm
    cube_default_xy_step = 0.035     # "
    cube_default_cunit = 'micron'

    def __init__(self, psf=None, lsf=None, lsf_fwhm=0.00065, psf_fwhm=0.091, psf_pa=0., psf_ba=1.0):

        if lsf is None:
            lsf = GaussianLineSpreadFunction(fwhm=lsf_fwhm)
        if psf is None:
            psf = GaussianPointSpreadFunction(fwhm=psf_fwhm, pa=psf_pa, ba=psf_ba)

        Instrument.__init__(self, psf=psf, lsf=lsf)

    def use_pixelsize_from_cube(self, cube):
        self.lsf.fwhm = 2.5 * cube.z_step
        Instrument.use_pixelsize_from_cube(self, cube)
